/*
 * Atheros CARL9170 driver
 *
 * Driver specific definitions
 *
 * Copyright 2008, Johannes Berg <johannes@sipsolutions.net>
 * Copyright 2009, Christian Lamparter <chunkeey@googlemail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, see
 * http://www.gnu.org/licenses/.
 *
 * This file incorporates work covered by the following copyright and
 * permission notice:
 *    Copyright (c) 2007-2008 Atheros Communications, Inc.
 *
 *    Permission to use, copy, modify, and/or distribute this software for any
 *    purpose with or without fee is hereby granted, provided that the above
 *    copyright notice and this permission notice appear in all copies.
 *
 *    THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 *    WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 *    MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 *    ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 *    WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 *    ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 *    OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifndef __CARL9170_H
#define __CARL9170_H

#include <linux/completion.h>
#include <linux/firmware.h>
#include <linux/spinlock.h>
#include <net/cfg80211.h>
#include <linux/kernel.h>
#include <net/mac80211.h>
#ifdef CONFIG_CARL9170_LEDS
#include <linux/leds.h>
#endif /* CONFIG_CARL170_LEDS */
#ifdef CONFIG_CARL9170_WPS_BUTTON
#include <linux/input-polldev.h>
#endif /* CONFIG_CARL9170_WPS_BUTTON */
#include "eeprom.h"
#include "wlan.h"
#include "hw.h"
#include "fwhdr.h"
#include "fwcmd.h"
#include "rc.h"
#include "../regd.h"

#ifdef CONFIG_CARL9170_DEBUGFS
#include "debug.h"
#endif /* CONFIG_CARL9170_DEBUGFS */

#define PAYLOAD_MAX	(CARL9170_MAX_CMD_LEN/4 - 1)

enum ar9170_bw {
	AR9170_BW_20,
	AR9170_BW_40_BELOW,
	AR9170_BW_40_ABOVE,

	__AR9170_NUM_BW,
};

static inline enum ar9170_bw nl80211_to_ar9170(enum nl80211_channel_type type)
{
	switch (type) {
	case NL80211_CHAN_NO_HT:
	case NL80211_CHAN_HT20:
		return AR9170_BW_20;
	case NL80211_CHAN_HT40MINUS:
		return AR9170_BW_40_BELOW;
	case NL80211_CHAN_HT40PLUS:
		return AR9170_BW_40_ABOVE;
	default:
		BUG();
	}
}

enum ar9170_rf_init_mode {
	AR9170_RFI_NONE,
	AR9170_RFI_WARM,
	AR9170_RFI_COLD,
};

#define AR9170_MAX_RX_BUFFER_SIZE		8192

#ifdef CONFIG_CARL9170_LEDS
struct ar9170;

struct ar9170_led {
	struct ar9170 *ar;
	struct led_classdev l;
	char name[32];
	unsigned int toggled;
	bool last_state;
	bool registered;
};

#endif /* CONFIG_CARL9170_LEDS */

enum ar9170_device_state {
	AR9170_UNKNOWN_STATE,
	AR9170_STOPPED,
	AR9170_IDLE,
	AR9170_STARTED,
};

struct ar9170_rxstream_mpdu_merge {
	struct ar9170_rx_head plcp;
	bool has_plcp;
};

#define AR9170_NUM_TID		16
#define WME_BA_BMP_SIZE         64

#define WME_AC_BE   2
#define WME_AC_BK   3
#define WME_AC_VI   1
#define WME_AC_VO   0

#define TID_TO_WME_AC(_tid)				\
	((((_tid) == 0) || ((_tid) == 3)) ? WME_AC_BE :	\
	 (((_tid) == 1) || ((_tid) == 2)) ? WME_AC_BK :	\
	 (((_tid) == 4) || ((_tid) == 5)) ? WME_AC_VI :	\
	 WME_AC_VO)


#define SEQ_DIFF(_start, _seq) \
	(((_start) - (_seq)) & 0x0fff)
#define SEQ_PREV(_seq) \
	(((_seq) - 1) & 0x0fff)
#define SEQ_NEXT(_seq) \
	(((_seq) + 1) & 0x0fff)
#define BAW_WITHIN(_start, _bawsz, _seqno) \
	((((_seqno) - (_start)) & 0xfff) < (_bawsz))

enum ar9170_tid_state {
	AR9170_TID_STATE_INVALID,
	AR9170_TID_STATE_SHUTDOWN,
	AR9170_TID_STATE_PROGRESS,
	AR9170_TID_STATE_IDLE,
	AR9170_TID_STATE_XMIT,
};

#define AR9170_BAW_BITS (2 * WME_BA_BMP_SIZE)
#define AR9170_BAW_SIZE (BITS_TO_LONGS(AR9170_BAW_BITS))
#define AR9170_BAW_LEN (DIV_ROUND_UP(AR9170_BAW_BITS, BITS_PER_BYTE))

struct ar9170_sta_tid {
	struct list_head list;

	/* temporary list for the RCU unlink procedure */
	struct list_head tmp_list;

	unsigned int counter;
	uint8_t tid;			/* simply TID number ( 0 - 15 ) */
	u16 max;			/* max. AMPDU size */

	/* lock for the following data structures */
	spinlock_t lock;

	u16 snx;	/* sequence of the _next_ frame xmit */
	u16 hsn;	/* sequence of the highest queued  */
	u16 bsn;	/* base of the tx/agg bitmap */
	unsigned long bitmap[AR9170_BAW_SIZE];

	/* Preaggregation reorder queue */
	struct sk_buff_head queue;

	enum ar9170_tid_state state;
};

#define AR9170_QUEUE_TIMEOUT		256
#define AR9170_BUMP_QUEUE		1000
#define AR9170_TX_TIMEOUT		5000
#define AR9170_JANITOR_DELAY		128
#define AR9170_QUEUE_STUCK_TIMEOUT	5500


#define AR9170_NUM_TX_AGG_MAX		30
#define AR9170_NUM_TX_LIMIT_HARD	(2 * AR9170_TXQ_DEPTH)
#define AR9170_NUM_TX_LIMIT_SOFT	(AR9170_TXQ_DEPTH - 8)

enum carl9170_firmware_types {
	CARL9170FW_V1,

	/* last entry */
	FIRMWARE_NUM
};

#define IS_CARL9170FW(a)						\
	(((struct ar9170 *)a)->firmware_type == CARL9170FW_V1)

struct ar9170 {
	struct ath_common common;
	struct ieee80211_hw *hw;
	struct device *pdev;
	struct mutex mutex;
	enum ar9170_device_state state;
	bool registered;
	unsigned long bad_hw_nagger;

	int (*open)(struct ar9170 *);
	void (*stop)(struct ar9170 *);
	int (*tx)(struct ar9170 *, struct sk_buff *);
	int (*exec_cmd)(struct ar9170 *, enum ar9170_cmd_oids, u32 ,
			void *, u32 , void *);
	void (*callback_cmd)(struct ar9170 *, u32 , void *);
	int (*flush)(struct ar9170 *);
	int (*restart)(struct ar9170 *, bool);

	/* firmware settings */
	enum carl9170_firmware_types firmware_type;
	bool shipped_fw_header_desc;
	const struct carl9170_fw_desc_head *fw_desc;

	/* reset / stuck frames/queue detection */
	unsigned int restart_counter;
	unsigned long queue_stop_timeout[__AR9170_NUM_TXQ];
	struct work_struct restart_work;
	bool needs_full_reset;

	/* interface mode settings */
	struct ieee80211_vif *vif;

	/* beaconing */
	struct sk_buff *beacon;
	struct work_struct beacon_work;
	u32 beacon_addr;
	u16 max_beacon_len;
	bool enable_beacon;

	/* cryptographic engine */
	u64 usedkeys;
	bool rx_software_decryption;
	bool disable_offload;

	/* filter settings */
	u64 cur_mc_hash;
	u32 cur_filter;
	unsigned int filter_state;
	bool sniffer_enabled;

	/* PHY */
	struct ieee80211_channel *channel;
	int noise[4];

	/* power calibration data */
	u8 power_5G_leg[4];
	u8 power_2G_cck[4];
	u8 power_2G_ofdm[4];
	u8 power_5G_ht20[8];
	u8 power_5G_ht40[8];
	u8 power_2G_ht20[8];
	u8 power_2G_ht40[8];

	u8 phy_heavy_clip;

#ifdef CONFIG_CARL9170_LEDS
	struct delayed_work led_work;
	struct ar9170_led leds[AR9170_NUM_LEDS];
#endif /* CONFIG_CARL9170_LEDS */

	/* qos queue settings */
	spinlock_t tx_stats_lock;
	struct ieee80211_tx_queue_stats tx_stats[__AR9170_NUM_TXQ];
	struct ieee80211_tx_queue_params edcf[5];

	spinlock_t cmdlock;
	__le32 cmdbuf[PAYLOAD_MAX + 1];

	/* MAC statistics */
	struct ieee80211_low_level_stats stats;

	/* EEPROM */
	struct ar9170_eeprom eeprom;

	/* tx queues - as seen by hw - */
	struct sk_buff_head tx_pending[__AR9170_NUM_TXQ];
	struct sk_buff_head tx_status[__AR9170_NUM_TXQ];
	struct delayed_work tx_janitor;
	unsigned long tx_janitor_last_run;

	/* tx ampdu */
	struct work_struct ampdu_work;
	spinlock_t tx_ampdu_list_lock;
	struct list_head tx_ampdu_list;
	unsigned int tx_ampdu_list_len;
	atomic_t tx_ampdu_pending;
	atomic_t tx_total_pending;
	atomic_t tx_total_queued;

	/* hw internal memory management */
	unsigned long *mem_cookie_jar;
	spinlock_t mem_jar_lock;
	atomic_t mem_free_blocks;
	unsigned int mem_blocks;
	unsigned int mem_block_size;
	atomic_t mem_allocs;

	/* rxstream mpdu merge */
	struct ar9170_rxstream_mpdu_merge rx_mpdu;
	struct sk_buff *rx_failover;
	int rx_failover_missing;

	/* rate control algorithm */
	const struct ath_rate_table *cur_rate_table;
	enum wireless_mode cur_rate_mode;
	bool use_prot;

#ifdef CONFIG_CARL9170_WPS_BUTTON
	bool has_wps_button;
	struct input_polled_dev *wps_pbc;
	char wps_name[32];
	char wps_phys[32];
#endif /* CONFIG_CARL9170_WPS_BUTTON */

#ifdef CONFIG_CARL9170_DEBUGFS
	struct ar9170_debug debug;
#endif /* CONFIG_CARL9170_DEBUGFS */
};

struct ar9170_ba_stats {
	uint8_t ampdu_len;
	uint8_t ampdu_ack_len;
	bool clear;
};

struct ar9170_sta_info {
	unsigned int ampdu_max_len;
	struct ar9170_sta_tid *agg[AR9170_NUM_TID];
	struct ar9170_ba_stats stats[AR9170_NUM_TID];
};

struct ar9170_tx_info {
	unsigned long timeout;
	struct ar9170 *ar;
	struct kref ref;
};

#define IS_STARTED(a)		(((struct ar9170 *)a)->state >= AR9170_STARTED)
#define IS_ACCEPTING_CMD(a)	(((struct ar9170 *)a)->state >= AR9170_IDLE)

/* exported interface */
void *ar9170_alloc(size_t priv_size);
int ar9170_register(struct ar9170 *ar);
void ar9170_rx(struct ar9170 *ar, struct sk_buff *skb);
void ar9170_unregister(struct ar9170 *ar);
void ar9170_tx_callback(struct ar9170 *ar, struct sk_buff *skb);
void ar9170_handle_command_response(struct ar9170 *ar, void *buf, u32 len);
int ar9170_nag_limiter(struct ar9170 *ar);

/* MAC */
int ar9170_init_mac(struct ar9170 *ar);
int ar9170_set_qos(struct ar9170 *ar);
int ar9170_update_multicast(struct ar9170 *ar, const u64 mc_hast);
int ar9170_update_frame_filter(struct ar9170 *ar, const u32 filter);
int ar9170_set_operating_mode(struct ar9170 *ar);
int ar9170_set_beacon_timers(struct ar9170 *ar);
int ar9170_set_dyn_sifs_ack(struct ar9170 *ar);
int ar9170_set_rts_cts_rate(struct ar9170 *ar);
int ar9170_set_ampdu_settings(struct ar9170 *ar);
int ar9170_set_slot_time(struct ar9170 *ar);
int ar9170_set_basic_rates(struct ar9170 *ar);
int ar9170_set_hwretry_limit(struct ar9170 *ar, u32 max_retry);
int ar9170_update_beacon(struct ar9170 *ar);
void ar9170_new_beacon(struct work_struct *work);
int ar9170_upload_key(struct ar9170 *ar, u8 id, const u8 *mac, u8 ktype,
		      u8 keyidx, u8 *keydata, int keylen);
int ar9170_disable_key(struct ar9170 *ar, u8 id);

/* TX */
int ar9170_op_tx(struct ieee80211_hw *hw, struct sk_buff *skb);
void ar9170_tx_janitor(struct work_struct *work);
void ar9170_tx_process_status(struct ar9170 *ar,
			      const struct ar9170_rsp *cmd);
void ar9170_tx_status(struct ar9170 *ar, struct sk_buff *skb,
		      const bool success);

/* LEDs */
#ifdef CONFIG_CARL9170_LEDS
int ar9170_register_leds(struct ar9170 *ar);
void ar9170_unregister_leds(struct ar9170 *ar);
#endif /* CONFIG_CARL9170_LEDS */
int ar9170_init_leds(struct ar9170 *ar);
int ar9170_set_leds_state(struct ar9170 *ar, u32 led_state);

/* PHY / RF */
int ar9170_init_phy(struct ar9170 *ar, enum ieee80211_band band);
int ar9170_init_rf(struct ar9170 *ar);
int ar9170_set_channel(struct ar9170 *ar, struct ieee80211_channel *channel,
		       enum ar9170_rf_init_mode rfi, enum ar9170_bw bw);

/* FW */
int ar9170_parse_firmware(struct ar9170 *ar, const struct firmware *fw);
const void *ar9170_fw_find_desc(struct ar9170 *ar, const u8 descid[4],
				const unsigned int len, const uint8_t min_ver);

void ar9170_debug_stat_rc(struct ar9170 *ar, int final_rate);
void ar9170_debug_stat_retries(struct ar9170 *ar, int rix,
			    int xretries, int retries, u8 per);

/* debug */
void carl9170_debug_stat_rc(struct ar9170 *ar, int final_rate);
void carl9170_debug_stat_retries(struct ar9170 *ar, int rix,
			    int xretries, int retries, u8 per);

int carl9170_rate_control_register(void);
void carl9170_rate_control_unregister(void);

extern struct ieee80211_rate __ar9170_ratetable[];
extern int modparam_ht;

static inline struct ieee80211_hdr *ar9170_get_hdr(struct sk_buff *skb)
{
	return (void *)((struct _carl9170_tx_superframe *)skb->data)->frame_data;
}

static inline u16 ar9170_get_seq_h(struct ieee80211_hdr *hdr)
{
	return le16_to_cpu(hdr->seq_ctrl) >> 4;
}

static inline u16 ar9170_get_seq(struct sk_buff *skb)
{
	return ar9170_get_seq_h(ar9170_get_hdr(skb));
}

static inline u16 ar9170_get_tid_h(struct ieee80211_hdr *hdr)
{
	return (ieee80211_get_qos_ctl(hdr))[0] & IEEE80211_QOS_CTL_TID_MASK;
}

static inline u16 ar9170_get_tid(struct sk_buff *skb)
{
	return ar9170_get_tid_h(ar9170_get_hdr(skb));
}

#endif /* __CARL9170_H */
