/*
 * Atheros CARL9170 driver
 *
 * debug(fs) probing
 *
 * Copyright 2008, Johannes Berg <johannes@sipsolutions.net>
 * Copyright 2009, Christian Lamparer <chunkeey@googlemail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, see
 * http://www.gnu.org/licenses/.
 *
 * This file incorporates work covered by the following copyright and
 * permission notice:
 *    Copyright (c) 2008-2009 Atheros Communications, Inc.
 *
 *    Permission to use, copy, modify, and/or distribute this software for any
 *    purpose with or without fee is hereby granted, provided that the above
 *    copyright notice and this permission notice appear in all copies.
 *
 *    THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 *    WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 *    MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 *    ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 *    WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 *    ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 *    OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <linux/seq_file.h>
#include "carl9170.h"
#include "usb.h"
#include "cmd.h"

#define ADD(buf, off, max, fmt, args...)				\
	off += snprintf(&buf[off], max - off, fmt, ##args);

static int ar9170_debugfs_open(struct inode *inode, struct file *file)
{
	file->private_data = inode->i_private;
	return 0;
}

struct ar9170_debugfs_fops {
	unsigned int read_bufsize;
	mode_t attr;
	char *(*read)(struct ar9170_usb *aru, char *buf, size_t bufsize,
		      ssize_t *len);
	ssize_t (*write)(struct ar9170_usb *aru, const char *buf, size_t size);
	const struct file_operations fops;
};

static ssize_t ar9170_debugfs_read(struct file *file,
				   char __user *userbuf,
				   size_t count, loff_t *ppos)
{
	struct ar9170_debugfs_fops *dfops;
	struct ar9170_usb *aru;
	char *buf = NULL, *res_buf = NULL;
	ssize_t ret = 0;
	int err = 0;

	if (!count)
		return 0;

	aru = file->private_data;

	if (!aru)
		return -ENODEV;
	dfops = container_of(file->f_op, struct ar9170_debugfs_fops, fops);

	if (!dfops->read) {
		err = -ENOSYS;
		goto out_unlock;
	}

	if (dfops->read_bufsize) {
		buf = vmalloc(dfops->read_bufsize);
		if (!buf) {
			err = -ENOMEM;
			goto out_unlock;
		}
	}

	mutex_lock(&aru->common.mutex);
	if (!IS_STARTED(&aru->common)) {
		err = -ENODEV;
		goto out_unlock;
	}

	res_buf = dfops->read(aru, buf, dfops->read_bufsize, &ret);

	if (ret > 0)
		err = simple_read_from_buffer(userbuf, count, ppos,
					      res_buf, ret);
	else
		err = ret;

	WARN_ON_ONCE(dfops->read_bufsize && (res_buf != buf));
	vfree(res_buf);

out_unlock:
	mutex_unlock(&aru->common.mutex);
	return err;
}

static ssize_t ar9170_debugfs_write(struct file *file,
				    const char __user *userbuf,
				    size_t count, loff_t *ppos)
{
	struct ar9170_debugfs_fops *dfops;
	struct ar9170_usb *aru;
	char *buf = NULL;
	int err = 0;

	if (!count)
		return 0;

	if (count > PAGE_SIZE)
		return -E2BIG;

	aru = file->private_data;

	if (!aru)
		return -ENODEV;
	dfops = container_of(file->f_op, struct ar9170_debugfs_fops, fops);

	buf = vmalloc(count);
	if (!buf) {
		err = -ENOMEM;
		goto out_unlock;
	}

	if (copy_from_user(buf, userbuf, count)) {
		err = -EFAULT;
		goto out_free;
	}

	mutex_lock(&aru->common.mutex);
	if (!IS_STARTED(&aru->common)) {
		err = -ENODEV;
		goto out_unlock;
	}

	if (!dfops->write) {
		err = -ENOSYS;
		goto out_unlock;
	}

	err = dfops->write(aru, buf, count);
	if (err)
		goto out_unlock;

out_unlock:
	mutex_unlock(&aru->common.mutex);

out_free:
	vfree(buf);
	return err;
}

#define DEBUGFS_DECLARE_FILE(name, _read, _write, _read_bufsize, _attr)	\
static const struct ar9170_debugfs_fops ar9170_debugfs_##name ##_ops = {\
	.read_bufsize = _read_bufsize,					\
	.read = _read,							\
	.write = _write,						\
	.attr = _attr,							\
	.fops = {							\
		.open	= ar9170_debugfs_open,				\
		.read	= ar9170_debugfs_read,				\
		.write	= ar9170_debugfs_write,				\
		.owner = THIS_MODULE 					\
	},								\
}

#define DEBUGFS_DECLARE_RO_FILE(name, _read_bufsize)			\
	DEBUGFS_DECLARE_FILE(name, ar9170_debugfs_##name ##_read, NULL, \
			     _read_bufsize, S_IRUSR)

#define DEBUGFS_DECLARE_WO_FILE(name)					\
	DEBUGFS_DECLARE_FILE(name, NULL, ar9170_debugfs_##name ##_write,\
			     0, S_IWUSR)

#define DEBUGFS_DECLARE_RW_FILE(name, _read_bufsize)			\
	DEBUGFS_DECLARE_FILE(name, ar9170_debugfs_##name ##_read,	\
			     ar9170_debugfs_##name ##_write,		\
			     _read_bufsize, S_IRUSR | S_IWUSR)


#define DEBUGFS_READONLY_FILE(name, _read_bufsize, fmt, value...)	\
static char *ar9170_debugfs_ ##name ## _read(struct ar9170_usb *aru,	\
					     char *buf, size_t buf_size,\
					     ssize_t *len)		\
{									\
	ADD(buf, *len, buf_size, fmt "\n", ##value); 			\
	return buf;							\
}									\
DEBUGFS_DECLARE_RO_FILE(name, _read_bufsize)

static char *ar9170_debugfs_mem_usage_read(struct ar9170_usb *aru,
					   char *buf, size_t bufsize,
					   ssize_t *len)
{
	struct ar9170 *ar = &aru->common;

	ADD(buf, *len, bufsize, "jar: [");

	spin_lock_irq(&ar->mem_jar_lock);

	*len += bitmap_scnprintf(&buf[*len], bufsize - *len,
				  ar->mem_cookie_jar, ar->mem_blocks);

	ADD(buf, *len, bufsize, "]\n");

	ADD(buf, *len, bufsize, "cookies: used:%3d / total:%3d, allocs:%d\n",
	    bitmap_weight(ar->mem_cookie_jar, ar->mem_blocks), ar->mem_blocks,
	    atomic_read(&ar->mem_allocs));

	ADD(buf, *len, bufsize, "memory: free:%3d (%3d KiB) / total:%3d KiB)\n",
	    atomic_read(&ar->mem_free_blocks),
	    (atomic_read(&ar->mem_free_blocks) * ar->mem_block_size) / 1024,
	    (ar->mem_blocks * ar->mem_block_size) / 1024);

	spin_unlock_irq(&ar->mem_jar_lock);

	return buf;
}
DEBUGFS_DECLARE_RO_FILE(mem_usage, 512);

static char *ar9170_debugfs_qos_stat_read(struct ar9170_usb *aru,
					  char *buf, size_t bufsize,
					  ssize_t *len)
{
	struct ar9170 *ar = &aru->common;

	ADD(buf, *len, bufsize, "%s QoS AC\n", modparam_ht ? "Software" : "Hardware");

	ADD(buf, *len, bufsize, "[     VO            VI            BE            BK      ]\n");

	spin_lock_irq(&ar->tx_stats_lock);
	ADD(buf, *len, bufsize, "[length/limit  length/limit  length/limit  length/limit ]\n"
				"[   %3d/%3d       %3d/%3d       %3d/%3d       %3d/%3d   ]\n\n",
	    ar->tx_stats[0].len, ar->tx_stats[0].limit,
	    ar->tx_stats[1].len, ar->tx_stats[1].limit,
	    ar->tx_stats[2].len, ar->tx_stats[2].limit,
	    ar->tx_stats[3].len, ar->tx_stats[3].limit);

	ADD(buf, *len, bufsize, "[    total         total         total         total     ]\n"
				"[%10d    %10d    %10d    %10d  ]\n\n",
	    ar->tx_stats[0].count, ar->tx_stats[1].count,
	    ar->tx_stats[2].count, ar->tx_stats[3].count);

	spin_unlock_irq(&ar->tx_stats_lock);

	ADD(buf, *len, bufsize, "[  pend/waittx   pend/waittx   pend/waittx   pend/waittx]\n"
				"[   %3d/%3d       %3d/%3d       %3d/%3d       %3d/%3d   ]\n\n",
	    skb_queue_len(&ar->tx_pending[0]), skb_queue_len(&ar->tx_status[0]),
	    skb_queue_len(&ar->tx_pending[1]), skb_queue_len(&ar->tx_status[1]),
	    skb_queue_len(&ar->tx_pending[2]), skb_queue_len(&ar->tx_status[2]),
	    skb_queue_len(&ar->tx_pending[3]), skb_queue_len(&ar->tx_status[3]));

	return buf;
}
DEBUGFS_DECLARE_RO_FILE(qos_stat, 512);

static void ar9170_debugfs_format_frame(struct ar9170 *ar,
					struct sk_buff *skb,
					const char *prefix,
					char *buf, ssize_t *off,
					ssize_t bufsize)
{
	struct _carl9170_tx_superframe *txc = (void *) skb->data;
	struct ieee80211_tx_info *txinfo = IEEE80211_SKB_CB(skb);
	struct ar9170_tx_info *arinfo = (void *) txinfo->rate_driver_data;
	struct ieee80211_hdr *hdr = (void *) txc->frame_data;

	ADD(buf, *off, bufsize, "%s %p, c:%2x, DA:%pM, sq:%4d, mc:%.4x, "
	    "pc:%.8x, to:%d ms\n", prefix, skb, txc->s.cookie,
	    ieee80211_get_DA(hdr), ar9170_get_seq_h(hdr),
	    le16_to_cpu(txc->f.mac_control), le32_to_cpu(txc->f.phy_control),
	    jiffies_to_msecs(jiffies - arinfo->timeout));
}


static char *ar9170_debugfs_ampdu_state_read(struct ar9170_usb *aru,
					     char *buf, size_t bufsize,
					     ssize_t *len)
{
	struct ar9170 *ar = &aru->common;
	struct ar9170_sta_tid *iter;
	struct sk_buff *skb;
	int cnt = 0, fc;
	int offset;

	rcu_read_lock();
	list_for_each_entry_rcu(iter, &ar->tx_ampdu_list, list) {

		spin_lock_irq(&iter->lock);
		ADD(buf, *len, bufsize, "Entry: #%2d TID:%1d, BSN:%4d, "
		    "SNX:%4d, HSN:%4d, Max BAW Size:%2d, state:%d, kicks:%d\n",
		    cnt, iter->tid, iter->bsn, iter->snx, iter->hsn,
		    iter->max, iter->state, iter->counter);

		ADD(buf, *len, bufsize, "\tWindow:  [");

		*len += bitmap_scnprintf(&buf[*len], bufsize - *len,
			iter->bitmap, AR9170_BAW_BITS);

#define BM_STR_OFF(offset) 						\
	((AR9170_BAW_BITS - (offset) - 1) / 4 +				\
	 (AR9170_BAW_BITS - (offset) - 1) / 32 + 1)

		ADD(buf, *len, bufsize, ",W]\n");

		offset = BM_STR_OFF(0);
		ADD(buf, *len, bufsize, "\tBase Seq: %*s\n", offset, "T");

		offset = BM_STR_OFF(SEQ_DIFF(iter->snx, iter->bsn));
		ADD(buf, *len, bufsize, "\tNext Seq: %*s\n", offset, "W");

		offset = BM_STR_OFF((int)iter->hsn - (int)iter->bsn);
		ADD(buf, *len, bufsize, "\tLast Seq: %*s\n", offset, "N");

		ADD(buf, *len, bufsize, "\tPre-Aggregation reorder buffer: "
		    " currently queued:%d\n", skb_queue_len(&iter->queue));

		fc = 0;
		skb_queue_walk(&iter->queue, skb) {
			char prefix[32];

			snprintf(prefix, sizeof(prefix), "\t\t%3d :", fc);
			ar9170_debugfs_format_frame(ar, skb, prefix,
						    buf, len, bufsize);

			fc++;
		}
		spin_unlock_irq(&iter->lock);
		cnt++;
	}
	rcu_read_unlock();

	return buf;
}
DEBUGFS_DECLARE_RO_FILE(ampdu_state, 8000);

static void ar9170_debugfs_queue_dump(struct ar9170 *ar, char *buf,
				      ssize_t *len, size_t bufsize,
				      struct sk_buff_head *queue)
{
	struct sk_buff *skb;
	char prefix[16];
	int fc = 0;

	spin_lock_irq(&queue->lock);
	skb_queue_walk(queue, skb) {

		snprintf(prefix, sizeof(prefix), "%3d :", fc);
		ar9170_debugfs_format_frame(ar, skb, prefix, buf,
					    len, bufsize);
		fc++;
	}
	spin_unlock_irq(&queue->lock);
}

#define DEBUGFS_QUEUE_DUMP(q, qi)					\
static char *ar9170_debugfs_##q ##_##qi ##_read(struct ar9170_usb *aru,	\
					    char *buf, size_t bufsize,	\
					    ssize_t *len)		\
{									\
	struct ar9170 *ar = &aru->common;				\
									\
	ar9170_debugfs_queue_dump(ar, buf, len, bufsize, &ar->q[qi]);	\
	return buf;							\
}									\
DEBUGFS_DECLARE_RO_FILE(q##_##qi, 8000);

void carl9170_debug_stat_rc(struct ar9170 *ar, int final_rate)
{
	struct ath_rc_stats *stats;

	stats = &ar->debug.stats.rcstats[final_rate];
	stats->success++;
}

void carl9170_debug_stat_retries(struct ar9170 *ar, int rix,
				 int xretries, int retries, u8 per)
{
	struct ath_rc_stats *stats = &ar->debug.stats.rcstats[rix];

	stats->xretries += xretries;
	stats->retries += retries;
	stats->per = per;
}


static char *ar9170_debugfs_rc_stats_read(struct ar9170_usb *aru,
					  char *dummy, size_t bufsize,
					  ssize_t *len)
{
	struct ar9170 *ar = &aru->common;
	char *buf;
	unsigned int max;
	int i = 0;

	if (!ar->cur_rate_table)
		return NULL;

	max = 80 + ar->cur_rate_table->rate_cnt * 64;
	buf = vmalloc(max);
	if (!buf)
		return NULL;

	ADD(buf, *len, max, "%5s %15s %8s %9s %3s\n\n", "Rate", "Success",
	    "Retries", "XRetries", "PER");

	for (i = 0; i < ar->cur_rate_table->rate_cnt; i++) {
		u32 ratekbps = ar->cur_rate_table->info[i].ratekbps;
		struct ath_rc_stats *stats = &ar->debug.stats.rcstats[i];

		ADD(buf, *len, max, "%3u.%d: %8u %8u %8u %8u\n",
		    ratekbps / 1000, (ratekbps % 1000) / 100, stats->success,
		    stats->retries, stats->xretries, stats->per);
	}

	return buf;
}
DEBUGFS_DECLARE_RO_FILE(rc_stats, 0);

#define UPDATE_COUNTER(ar, name)	({				\
	u32 __tmp[ARRAY_SIZE(name##_regs)];				\
	unsigned int __i, __err = -ENODEV;				\
									\
	for (__i = 0; __i < ARRAY_SIZE(name##_regs); __i++) {		\
		__tmp[__i] = name##_regs[__i].reg;			\
		ar->debug.stats.name##_counter[__i] = 0;		\
	}								\
									\
	if (IS_STARTED(ar))						\
		__err = ar9170_read_mreg(ar, ARRAY_SIZE(name##_regs),	\
			__tmp, ar->debug.stats.name##_counter);		\
	(__err); })

#define TALLY_SUM_UP(ar, name)	do {					\
	unsigned int __i;						\
									\
	for (__i = 0; __i < ARRAY_SIZE(name##_regs); __i++) {		\
		ar->debug.stats.name##_sum[__i] +=			\
			ar->debug.stats.name##_counter[__i];		\
	}								\
} while (0)

#define DEBUGFS_HW_TALLY_FILE(name, f)					\
static char *ar9170_debugfs_##name ## _read(struct ar9170_usb *aru,	\
					    char *dum, size_t bufsize,\
					    ssize_t *ret)		\
{									\
	struct ar9170 *ar = &aru->common;				\
	char *buf;							\
	int i, max_len, err;						\
									\
	max_len = ARRAY_SIZE(name##_regs) * 80;				\
	buf = vmalloc(max_len);						\
	if (!buf)							\
		return NULL;						\
									\
	err = UPDATE_COUNTER(ar, name);					\
	if (err) {							\
		*ret = err;						\
		return buf;						\
	}								\
									\
	TALLY_SUM_UP(ar, name);						\
									\
	for (i = 0; i < ARRAY_SIZE(name##_regs); i++) {			\
		ADD(buf, *ret, max_len, "%22s = %" f "[+%" f "]\n",	\
		    name##_regs[i].nreg, ar->debug.stats.name ##_sum[i],\
		    ar->debug.stats.name ##_counter[i]);		\
	}								\
									\
	return buf;							\
}									\
DEBUGFS_DECLARE_RO_FILE(name, 0);

#define DEBUGFS_HW_REG_FILE(name, f)					\
static char *ar9170_debugfs_##name ## _read(struct ar9170_usb *aru,	\
					    char *dum, size_t bufsize,	\
					    ssize_t *ret)		\
{									\
	struct ar9170 *ar = &aru->common;				\
	char *buf;							\
	int i, max_len, err;						\
									\
	max_len = ARRAY_SIZE(name##_regs) * 80;				\
	buf = vmalloc(max_len);						\
	if (!buf)							\
		return NULL;						\
									\
	err = UPDATE_COUNTER(ar, name);					\
	if (err) {							\
		*ret = err;						\
		return buf;						\
	}								\
									\
	for (i = 0; i < ARRAY_SIZE(name##_regs); i++) {			\
		ADD(buf, *ret, max_len, "%22s = %" f "\n",		\
		    name##_regs[i].nreg,				\
		    ar->debug.stats.name##_counter[i]);			\
	}								\
									\
	return buf;							\
}									\
DEBUGFS_DECLARE_RO_FILE(name, 0);

static ssize_t ar9170_debugfs_hw_ioread32_write(struct ar9170_usb *aru,
						const char *buf, size_t count)
{
	struct ar9170 *ar = &aru->common;
	int err = 0, i, n = 0, max_len = 32, res;
	unsigned int reg, tmp;

	if (!count)
		return 0;

	if (count > max_len)
		return -E2BIG;

	res = sscanf(buf, "0x%X %d", &reg, &n);
	if (res < 1) {
		err = -EINVAL;
		goto out;
	}

	if (res == 1)
		n = 1;

	if (n > 15) {
		err = -EMSGSIZE;
		goto out;
	}

	if ((reg >= 0x280000) || ((reg + (n << 2)) >= 0x280000)) {
		err = -EADDRNOTAVAIL;
		goto out;
	}

	if (reg & 3) {
		err = -EINVAL;
		goto out;
	}

	for (i = 0; i < n; i++) {
		err = ar9170_read_reg(ar, reg + (i << 2), &tmp);
		if (err)
			goto out;

		ar->debug.ring[ar->debug.ring_tail].reg = reg + (i << 2);
		ar->debug.ring[ar->debug.ring_tail].value = tmp;
		ar->debug.ring_tail++;
		ar->debug.ring_tail %= AR9170_DEBUG_RING_SIZE;
	}

out:
	return err ? err : count;
}

static char *ar9170_debugfs_hw_ioread32_read(struct ar9170_usb *aru,
					     char *buf, size_t bufsize,
					     ssize_t *ret)
{
	struct ar9170 *ar = &aru->common;
	int i = 0;

	while (ar->debug.ring_head != ar->debug.ring_tail) {
		ADD(buf, *ret, bufsize, "%.8x = %.8x\n",
		    ar->debug.ring[ar->debug.ring_head].reg,
		    ar->debug.ring[ar->debug.ring_head].value);

		ar->debug.ring_head++;
		ar->debug.ring_head %= AR9170_DEBUG_RING_SIZE;

		if (i++ == 64)
			break;
	}
	ar->debug.ring_head = ar->debug.ring_tail;
	return buf;
}
DEBUGFS_DECLARE_RW_FILE(hw_ioread32, AR9170_DEBUG_RING_SIZE * 40);

static ssize_t ar9170_debugfs_restart_write(struct ar9170_usb *aru,
					    const char *buf, size_t count)
{
	struct ar9170 *ar = &aru->common;

	if (count < 1)
		return -EINVAL;

	switch (buf[0]) {
	case 'F':
		ar->needs_full_reset = true;
		break;

	case 'R':
		ar->needs_full_reset = false;
		break;

	case 'M':
		ar9170_mac_reset(ar);
		goto out;

	default:
		return -EINVAL;
	}

	ieee80211_queue_work(ar->hw, &ar->restart_work);

out:
	return count;
}

static char *ar9170_debugfs_restart_read(struct ar9170_usb *aru,
					     char *buf, size_t bufsize,
					     ssize_t *ret)
{
	ADD(buf, *ret, bufsize, "firmware restarts:%d\n",
		aru->common.restart_counter);
	return buf;
}
DEBUGFS_DECLARE_RW_FILE(restart, 64);

static ssize_t ar9170_debugfs_hw_iowrite32_write(struct ar9170_usb *aru,
						 const char *buf, size_t count)
{
	struct ar9170 *ar = &aru->common;
	int err = 0, max_len = 22, res;
	u32 reg, val;

	if (!count)
		return 0;

	if (count > max_len)
		return -E2BIG;

	res = sscanf(buf, "0x%X 0x%X", &reg, &val);
	if (res != 2) {
		err = -EINVAL;
		goto out;
	}

	if (reg <= 0x100000 || reg >= 0x280000) {
		err = -EADDRNOTAVAIL;
		goto out;
	}

	if (reg & 3) {
		err = -EINVAL;
		goto out;
	}

	err = ar9170_write_reg(ar, reg, val);
	if (err)
		goto out;

out:
	return err ? err : count;
}
DEBUGFS_DECLARE_WO_FILE(hw_iowrite32);

DEBUGFS_HW_TALLY_FILE(hw_tx_tally, "u");
DEBUGFS_HW_TALLY_FILE(hw_rx_tally, "u");
DEBUGFS_HW_TALLY_FILE(hw_phy_errors, "u");
DEBUGFS_HW_REG_FILE(hw_wlan_queue, ".8x");
DEBUGFS_HW_REG_FILE(hw_pta_queue, ".8x");
DEBUGFS_HW_REG_FILE(hw_ampdu_info, ".8x");
DEBUGFS_QUEUE_DUMP(tx_status, 0);
DEBUGFS_QUEUE_DUMP(tx_status, 1);
DEBUGFS_QUEUE_DUMP(tx_status, 2);
DEBUGFS_QUEUE_DUMP(tx_status, 3);
DEBUGFS_QUEUE_DUMP(tx_pending, 0);
DEBUGFS_QUEUE_DUMP(tx_pending, 1);
DEBUGFS_QUEUE_DUMP(tx_pending, 2);
DEBUGFS_QUEUE_DUMP(tx_pending, 3);
DEBUGFS_READONLY_FILE(usb_submitted_urbs, 20, "%d",
	      atomic_read(&aru->tx_submitted_urbs));

DEBUGFS_READONLY_FILE(usb_pending_urbs, 20, "%d",
		      aru->tx_pending_urbs);

DEBUGFS_READONLY_FILE(tx_total_queued, 20, "%d",
		      atomic_read(&aru->common.tx_total_queued));

DEBUGFS_READONLY_FILE(tx_total_pending, 20, "%d",
		      atomic_read(&aru->common.tx_total_pending));

DEBUGFS_READONLY_FILE(tx_ampdu_list_len, 20, "%d",
		      aru->common.tx_ampdu_list_len);

DEBUGFS_READONLY_FILE(tx_ampdu_pending, 20, "%d",
		      atomic_read(&aru->common.tx_ampdu_pending));

DEBUGFS_READONLY_FILE(tx_janitor_last_run, 64, "last run:%d ms ago",
	jiffies_to_msecs(jiffies - aru->common.tx_janitor_last_run));

DEBUGFS_READONLY_FILE(firmware_type, 20, "%d",
		      aru->common.firmware_type);

DEBUGFS_READONLY_FILE(restart_count, 20, "%d",
		      aru->common.restart_counter);

void ar9170_register_debugfs(struct ar9170_usb *aru)
{
	aru->debug_dir = debugfs_create_dir(KBUILD_MODNAME,
		aru->common.hw->wiphy->debugfsdir);

#define DEBUGFS_ADD(name)						\
	debugfs_create_file(#name, ar9170_debugfs_##name ##_ops.attr,	\
			    aru->debug_dir, aru,			\
			    &ar9170_debugfs_##name ## _ops.fops);

	DEBUGFS_ADD(usb_submitted_urbs);
	DEBUGFS_ADD(usb_pending_urbs);

	DEBUGFS_ADD(tx_total_queued);
	DEBUGFS_ADD(tx_total_pending);
	DEBUGFS_ADD(tx_ampdu_pending);
	DEBUGFS_ADD(tx_ampdu_list_len);

	DEBUGFS_ADD(firmware_type);

	DEBUGFS_ADD(mem_usage);
	DEBUGFS_ADD(qos_stat);
	DEBUGFS_ADD(ampdu_state);

	DEBUGFS_ADD(hw_tx_tally);
	DEBUGFS_ADD(hw_rx_tally);
	DEBUGFS_ADD(hw_phy_errors);

	DEBUGFS_ADD(hw_wlan_queue);
	DEBUGFS_ADD(hw_pta_queue);
	DEBUGFS_ADD(hw_ampdu_info);

	DEBUGFS_ADD(tx_janitor_last_run);

	DEBUGFS_ADD(tx_status_0);
	DEBUGFS_ADD(tx_status_1);
	DEBUGFS_ADD(tx_status_2);
	DEBUGFS_ADD(tx_status_3);

	DEBUGFS_ADD(tx_pending_0);
	DEBUGFS_ADD(tx_pending_1);
	DEBUGFS_ADD(tx_pending_2);
	DEBUGFS_ADD(tx_pending_3);

	DEBUGFS_ADD(hw_ioread32);
	DEBUGFS_ADD(hw_iowrite32);
	DEBUGFS_ADD(restart);

	if (modparam_ht)
		DEBUGFS_ADD(rc_stats);

#undef DEBUGFS_ADD
}

void ar9170_unregister_debugfs(struct ar9170_usb *aru)
{
	debugfs_remove_recursive(aru->debug_dir);
}
