/*
 * Atheros CARL9170 driver
 *
 * firmware (parsing)
 *
 * Copyright 2009, Christian Lamparter <chunkeey@googlemail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, see
 * http://www.gnu.org/licenses/.
 */

#include <linux/kernel.h>
#include <linux/firmware.h>
#include <linux/crc32.h>
#include "carl9170.h"
#include "fwcmd.h"

static const u8 otus_magic[4] = { OTUS_MAGIC };
static const u8 last_magic[4] = { LAST_MAGIC };

const void *ar9170_fw_find_desc(struct ar9170 *ar,
				const u8 descid[4],
				const unsigned int len,
				const u8 compatible_revision)
{
	const struct carl9170_fw_desc_head *iter, *ret = NULL;

	for_each_hdr(iter, ar->fw_desc) {
		if (!memcmp(iter->magic, descid, sizeof(*descid)) &&
		    !CHECK_HDR_VERSION(iter, compatible_revision) &&
		    (iter->length >= len)) {
			ret = (void *)iter;
			break;
		}
	}

	return ret;
}

static int ar9170_fw_verify_descs(struct ar9170 *ar,
				  const struct carl9170_fw_desc_head *head,
				  unsigned int max_len)
{
	const struct carl9170_fw_desc_head *pos, *end;

	if (max_len < (sizeof(*pos)))
		return -ENODATA;

	max_len = min_t(unsigned int, CARL9170_FW_DESC_MAX_LENGTH, max_len);

	pos = head;
	end = (void *) head + max_len;

	while (pos < end) {
		if (((unsigned long)pos + pos->length) > (unsigned long)end)
			return -EMSGSIZE;

		if (pos->length > max_len)
			return -EMSGSIZE;

		if (pos->length < sizeof(struct carl9170_fw_desc_head))
			return -EINVAL;

		if (!memcmp(pos->magic, last_magic, sizeof(last_magic)))
			return 0;

		pos = (void *)((unsigned long)pos + pos->length);
		max_len -= pos->length;
	}

	return -EINVAL;
}

static int ar9170_fw_check(struct ar9170 *ar, const __u8 *data, size_t len)
{
	const struct carl9170_fw_otus_desc_v1 *otus_desc;

	otus_desc = (void *) ar9170_fw_find_desc(ar, otus_magic,
		sizeof(*otus_desc), CARL9170_FW_OTUS_DESC_CUR_VER);
	if (!otus_desc)
		return -ENODATA;

#define SUPP(feat)						\
	(is_feature_supported(otus_desc->fw_feature_set, feat))

	if (!SUPP(CARL9170_FW_DUMMY_FEATURE)) {
		dev_err(ar->pdev, "invalid firmware header format detected\n");
		return -EINVAL;
	}

	if (!SUPP(AR9170_FW_COMMAND_BASIC) || !SUPP(AR9170_FW_COMMAND_PHY) ||
	    SUPP(CARL9170_FW_UNUSEABLE) || !SUPP(AR9170_FW_HANDLE_BACK_REQ)) {
		dev_err(ar->pdev, "Firmware does not provide "
			"the minimal feature list.\n");
		return -ECANCELED;
	}

	if (!SUPP(CARL9170_FW_INTEGRATED_DESC_HEADER) &&
	    !ar->shipped_fw_header_desc) {
		dev_err(ar->pdev, "driver does not know how to find the "
			"real firmware offset.\n");
		return -EOPNOTSUPP;
	}

	if (!SUPP(AR9170_FW_COMMAND_CAM)) {
		dev_info(ar->pdev, "crypto offloading disabled by firmware.\n");
		ar->disable_offload = true;
	}

	if (SUPP(CARL9170_FW_CARL9170_V1)) {
		u16 fw_date = le16_to_cpu(otus_desc->fw_year_month_day);

		dev_info(ar->pdev, "firmware date: 2%03d-%02d-%02d\n",
			 CARL9170_FW_GET_YEAR(fw_date),
			 CARL9170_FW_GET_MONTH(fw_date),
			 CARL9170_FW_GET_DAY(fw_date));
	} else {
		dev_err(ar->pdev, "unsupported firmware.\n");
		return -EINVAL;
	}

	ar->beacon_addr = le32_to_cpu(otus_desc->bcn_addr);
	ar->max_beacon_len = le16_to_cpu(otus_desc->bcn_len);

	if (SUPP(CARL9170_FW_WLANTX_CAB))
		ar->hw->wiphy->interface_modes |= BIT(NL80211_IFTYPE_AP);

	if (SUPP(CARL9170_FW_CRC32_CHECKSUM_TAIL)) {
		__le32 crc32;

		crc32 = crc32_le(~0, data, len - 4);
		if (!memcmp(&crc32, &data[len - 4], 4) == 0) {
			dev_err(ar->pdev, "CRC32 checksum mismatch!\n");
			return -EINVAL;
		}
	}

#undef SUPPORTED
	return 0;
}

static struct carl9170_fw_desc_head *
ar9170_find_fw_desc_entry(struct ar9170 *ar, const size_t len,
			  const __u8 *fw_data)
{
	int scan = 0, found = 0;

	if (len < 1024 || len > 32768) {
		dev_err(ar->pdev, "firmware size is out of bound.\n");
		return NULL;
	}

	while (scan < len - 512) {
		if (fw_data[scan++] == otus_magic[found])
			found++;
		else
			found = 0;

		if (scan >= len)
			break;

		if (found == sizeof(otus_magic))
			break;
	}

	if (found != sizeof(otus_magic)) {
		return NULL;
	}

	return (void *) &fw_data[scan - found];
}

int ar9170_parse_firmware(struct ar9170 *ar, const struct firmware *fw)
{
	const struct carl9170_fw_desc_head *fw_desc = NULL;
	const __u8 *fw_data;
	unsigned long header_offset = 0;
	int err;

	fw_data = (const __u8 *) fw->data;

	fw_desc = ar9170_find_fw_desc_entry(ar, fw->size, fw_data);

	if (IS_CARL9170FW(ar))
		header_offset = (unsigned long)fw_desc - (unsigned long)fw_data;

	if (!fw_desc) {
		dev_err(ar->pdev, "unsupported firmware.\n");
		return -ENODATA;
	}

	err = ar9170_fw_verify_descs(ar, fw_desc, fw->size - header_offset);
	if (err) {
		dev_err(ar->pdev, "damaged firmware descriptor detected.\n");
		return err;
	}

	ar->fw_desc = fw_desc;

	err = ar9170_fw_check(ar, fw_data, fw->size);
	if (err) {
		dev_err(ar->pdev, "unsupported firmware.\n");
		return err;
	}

	return 0;
}
