/*
 * Atheros CARL9170 driver
 *
 * 802.11 xmit & status routines
 *
 * Copyright 2008, Johannes Berg <johannes@sipsolutions.net>
 * Copyright 2009, Christian Lamparter <chunkeey@web.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, see
 * http://www.gnu.org/licenses/.
 *
 * This file incorporates work covered by the following copyright and
 * permission notice:
 *    Copyright (c) 2007-2008 Atheros Communications, Inc.
 *
 *    Permission to use, copy, modify, and/or distribute this software for any
 *    purpose with or without fee is hereby granted, provided that the above
 *    copyright notice and this permission notice appear in all copies.
 *
 *    THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 *    WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 *    MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 *    ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 *    WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 *    ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 *    OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/etherdevice.h>
#include <net/mac80211.h>
#include "carl9170.h"
#include "hw.h"
#include "cmd.h"

static void ar9170_tx(struct ar9170 *ar);
static void ar9170_tx_ampdu(struct ar9170 *ar);

static inline unsigned int __ar9170_get_queue(struct ar9170 *ar,
					      const unsigned int queue)
{
	if (modparam_ht) {
		/*
		 * This is _hopefully_ just a temporary workaround
		 * until someone figures out how to get QoS and
		 * AMPDU to play nicely together.
		 */

		return 2;		/* AC_BE */
	} else {
		return queue;
	}
}

static inline unsigned int ar9170_get_queue(struct ar9170 *ar,
					    struct sk_buff *skb)
{
	return __ar9170_get_queue(ar, skb_get_queue_mapping(skb));
}

/*
 * Debug Definitions.
 *
 * #define AR9170_QUEUE_STOP_DEBUG
 * #define AR9170_QUEUE_DEBUG
 * #define AR9170_TXAGG_DEBUG
 */

#if (defined AR9170_QUEUE_DEBUG) || (defined AR9170_TXAGG_DEBUG)
static void ar9170_print_txheader(struct ar9170 *ar, struct sk_buff *skb)
{
	struct _carl9170_tx_superframe *txc = (void *) skb->data;
	struct ieee80211_tx_info *txinfo = IEEE80211_SKB_CB(skb);
	struct ar9170_tx_info *arinfo = (void *) txinfo->rate_driver_data;
	struct ieee80211_hdr *hdr = (void *) txc->frame_data;

	printk(KERN_DEBUG "%s: => FRAME [skb:%p, c:%x q:%d, DA:[%pM] s:%d "
			  "mac_ctrl:%04x, phy_ctrl:%08x, timeout:[%d ms]]\n",
	       wiphy_name(ar->hw->wiphy), skb, txc->s.cookie,
	       ar9170_get_queue(ar, skb), ieee80211_get_DA(hdr),
	       ar9170_get_seq_h(hdr), le16_to_cpu(txc->f.mac_control),
	       le32_to_cpu(txc->f.phy_control),
	       jiffies_to_msecs(jiffies - arinfo->timeout));
}

static void __ar9170_dump_txqueue(struct ar9170 *ar,
				struct sk_buff_head *queue)
{
	struct sk_buff *skb;
	int i = 0;

	printk(KERN_DEBUG "---[ cut here ]---\n");
	printk(KERN_DEBUG "%s: %d entries in queue.\n",
	       wiphy_name(ar->hw->wiphy), skb_queue_len(queue));

	skb_queue_walk(queue, skb) {
		printk(KERN_DEBUG "index:%d => \n", i++);
		ar9170_print_txheader(ar, skb);
	}
	if (i != skb_queue_len(queue))
		printk(KERN_DEBUG "WARNING: queue frame counter "
		       "mismatch %d != %d\n", skb_queue_len(queue), i);
	printk(KERN_DEBUG "---[ end ]---\n");
}
#endif /* AR9170_QUEUE_DEBUG || AR9170_TXAGG_DEBUG */

#ifdef AR9170_QUEUE_DEBUG
static void ar9170_dump_txqueue(struct ar9170 *ar,
				struct sk_buff_head *queue)
{
	unsigned long flags;

	spin_lock_irqsave(&queue->lock, flags);
	__ar9170_dump_txqueue(ar, queue);
	spin_unlock_irqrestore(&queue->lock, flags);
}
#endif /* AR9170_QUEUE_DEBUG */

static void ar9170_tx_accounting(struct ar9170 *ar, struct sk_buff *skb)
{
	unsigned long flags;
	int queue;

	atomic_inc(&ar->tx_total_queued);

	queue = skb_get_queue_mapping(skb);
	spin_lock_irqsave(&ar->tx_stats_lock, flags);

	/*
	 * This frame is going to be queued, regardless if the queue is
	 * full to the brim, or not. We have to do the queuing internally,
	 * since mac80211 assumes that a driver which can operate with
	 * aggregated frames does not reject frames for this reason.
	 */
	ar->tx_stats[queue].len++;
	ar->tx_stats[queue].count++;

	if (ar->tx_stats[queue].len >= ar->tx_stats[queue].limit) {
#ifdef AR9170_QUEUE_DEBUG
		printk(KERN_DEBUG "%s: queue %d full\n",
		       wiphy_name(ar->hw->wiphy), queue);

		printk(KERN_DEBUG "%s: stuck frames: ===> \n",
		       wiphy_name(ar->hw->wiphy));
		ar9170_dump_txqueue(ar, &ar->tx_pending[queue]);
		ar9170_dump_txqueue(ar, &ar->tx_status[queue]);
#endif /* AR9170_QUEUE_DEBUG */

#ifdef AR9170_QUEUE_STOP_DEBUG
		printk(KERN_DEBUG "%s: stop queue %d\n",
		       wiphy_name(ar->hw->wiphy), queue);
#endif /* AR9170_QUEUE_STOP_DEBUG */

		ieee80211_stop_queue(ar->hw, queue);
		ar->queue_stop_timeout[queue] = jiffies;
	}

	spin_unlock_irqrestore(&ar->tx_stats_lock, flags);
}

static void ar9170_tx_accounting_free(struct ar9170 *ar, struct sk_buff *skb)
{
	unsigned long flags;
	struct ieee80211_tx_info *txinfo;
	int queue, i;

	txinfo = IEEE80211_SKB_CB(skb);
	queue = skb_get_queue_mapping(skb);

	spin_lock_irqsave(&ar->tx_stats_lock, flags);

	ar->tx_stats[queue].len--;

	for (i = 0; i < ar->hw->queues; i++) {
		if (DIV_ROUND_UP(IEEE80211_MAX_FRAME_LEN, ar->mem_block_size) >
		    atomic_read(&ar->mem_free_blocks))
			break;

		if (ar->tx_stats[i].len < AR9170_NUM_TX_LIMIT_SOFT) {
#ifdef AR9170_QUEUE_STOP_DEBUG
			if (ieee80211_queue_stopped(ar->hw, queue)) {
				printk(KERN_DEBUG "%s: wake queue %d\n",
				       wiphy_name(ar->hw->wiphy), queue);
			}
#endif /* AR9170_QUEUE_STOP_DEBUG */

			ieee80211_wake_queue(ar->hw, i);
		}
	}

	spin_unlock_irqrestore(&ar->tx_stats_lock, flags);
	atomic_dec(&ar->tx_total_queued);
}

static int ar9170_alloc_dev_space(struct ar9170 *ar, unsigned int len)
{
	unsigned long flags;
	unsigned int chunks;
	int cookie = -1;

	atomic_inc(&ar->mem_allocs);

	chunks = DIV_ROUND_UP(len, ar->mem_block_size);
	if (unlikely(atomic_sub_return(chunks, &ar->mem_free_blocks) < 6)) {
		atomic_add(chunks, &ar->mem_free_blocks);
		goto out;
	}

	spin_lock_irqsave(&ar->mem_jar_lock, flags);
	/* Put a check on the untamable hunger of the cookie monster! */
	cookie = bitmap_find_free_region(ar->mem_cookie_jar,
					 ar->mem_blocks, 0);

	spin_unlock_irqrestore(&ar->mem_jar_lock, flags);

	if (unlikely(cookie < 1))
		atomic_add(chunks, &ar->mem_free_blocks);

	if (WARN_ON_ONCE(cookie == 0))
		return -1;

out:
	return cookie;
}

static void ar9170_release_dev_space(struct ar9170 *ar, struct sk_buff *skb)
{
	struct _carl9170_tx_superframe *super = (void *) skb->data;
	unsigned long flags;

	/* cookie "0" is reserved and should never be freed */
	if (unlikely(WARN_ON_ONCE(!super->s.cookie)))
		return;

	atomic_add(DIV_ROUND_UP(skb->len, ar->mem_block_size),
		   &ar->mem_free_blocks);

	spin_lock_irqsave(&ar->mem_jar_lock, flags);
	bitmap_release_region(ar->mem_cookie_jar, super->s.cookie, 0);
	spin_unlock_irqrestore(&ar->mem_jar_lock, flags);
}

static void ar9170_tx_release(struct kref *ref)
{
	struct ar9170 *ar;
	struct ar9170_tx_info *arinfo;
	struct ieee80211_tx_info *txinfo;
	struct sk_buff *skb;

	arinfo = container_of(ref, struct ar9170_tx_info, ref);
	txinfo = container_of((void *) arinfo, struct ieee80211_tx_info, rate_driver_data);
	skb = container_of((void *) txinfo, struct sk_buff, cb);

	ar = arinfo->ar;
	if (WARN_ON_ONCE(!ar))
		return ;

	BUILD_BUG_ON(
	    offsetof(struct ieee80211_tx_info, status.ampdu_ack_len) != 23);

	memset(&txinfo->status.ampdu_ack_len, 0,
	       sizeof(struct ieee80211_tx_info) -
	       offsetof(struct ieee80211_tx_info, status.ampdu_ack_len));

	if (txinfo->flags & IEEE80211_TX_STAT_AMPDU) {
		txinfo->status.ampdu_len = txinfo->pad[0];
		txinfo->status.ampdu_ack_len = txinfo->pad[1];
		txinfo->pad[0] = txinfo->pad[1] = 0;
	}

	skb_pull(skb, sizeof(struct _carl9170_tx_superframe));
	ieee80211_tx_status_irqsafe(ar->hw, skb);
}

/* caller must hold the tid_info->lock & rcu_read_lock */
static void ar9170_tx_shift_bm(struct ar9170 *ar,
			       struct ar9170_sta_tid *tid_info,
			       const u16 seq)
{
	u16 off;

	off = SEQ_DIFF(seq, tid_info->bsn);

	if (unlikely(off >= AR9170_BAW_BITS)) {
#ifdef AR9170_TXAGG_DEBUG
		printk(KERN_DEBUG "%s: BAS:[t:%d] ovr. tx_status for "
		       "frame[seq:%4d] at bitmap pos:%2d.\n",
		       wiphy_name(ar->hw->wiphy), tid_info->tid, seq, off);
#endif /* AR9170_TXAGG_DEBUG */
		return ;
	}

	if (!test_and_clear_bit(off, tid_info->bitmap)) {
#ifdef AR9170_TXAGG_DEBUG
		printk(KERN_DEBUG "%s: BAS:[t:%d] received invalid "
		       "tx_status for frame[seq:%4d] at bitmap pos:%2d.\n",
		       wiphy_name(ar->hw->wiphy), tid_info->tid, seq, off);
#endif /* AR9170_TXAGG_DEBUG */
	}

	off = SEQ_DIFF(tid_info->snx, tid_info->bsn);
	if (likely(off < AR9170_BAW_BITS)) {
		if (!bitmap_empty(tid_info->bitmap, off))
			off = find_first_bit(tid_info->bitmap, off);

		tid_info->bsn += off;
		tid_info->bsn &= 0x0fff;

		bitmap_shift_right(tid_info->bitmap, tid_info->bitmap,
				   off, AR9170_BAW_BITS);
	}
}

static void ar9170_tx_status_process_ampdu(struct ar9170 *ar,
					   struct sk_buff *skb,
					   struct ieee80211_tx_info *txinfo)
{
	unsigned long flags;
	struct _carl9170_tx_superframe *super = (void *) skb->data;
	struct ieee80211_hdr *hdr = (void *) super->frame_data;
	struct ar9170_sta_info *sta_info;
	struct ieee80211_sta *sta;
	struct ar9170_sta_tid *tid_info;
	u8 tid;

	if (!(txinfo->flags & IEEE80211_TX_CTL_AMPDU))
		return;

	rcu_read_lock();

	/*
	 * Normally we should use wrappers like ieee80211_get_DA to get
	 * the correct peer ieee80211_sta.
	 *
	 * But there is a problem with indirect traffic (broadcasts, or
	 * data which is designated for other stations) in station mode.
	 * The frame will be directed to the AP for distribution and not
	 * the actual destination.
	 */
	sta = ieee80211_find_sta(ar->vif, hdr->addr1);
	if (unlikely(!sta))
		goto out_rcu;

	tid = ar9170_get_tid_h(hdr);

	rcu_read_lock();
	sta_info = (void *) sta->drv_priv;
	tid_info = rcu_dereference(sta_info->agg[tid]);
	if (unlikely(!tid_info)) {
		rcu_read_unlock();
		goto out_rcu;
	}

	spin_lock_irqsave(&tid_info->lock, flags);
	if (likely(tid_info->state >= AR9170_TID_STATE_IDLE))
		ar9170_tx_shift_bm(ar, tid_info, ar9170_get_seq_h(hdr));

	spin_unlock_irqrestore(&tid_info->lock, flags);

	if (sta_info->stats[tid].clear) {
		sta_info->stats[tid].clear = false;
		sta_info->stats[tid].ampdu_len = 0;
		sta_info->stats[tid].ampdu_ack_len = 0;
	}

	sta_info->stats[tid].ampdu_len++;
	if (txinfo->status.rates[0].count == 1 &&
	    txinfo->flags & IEEE80211_TX_STAT_ACK)
		sta_info->stats[tid].ampdu_ack_len++;

	if (super->f.mac_control & cpu_to_le16(AR9170_TX_MAC_IMM_BA)) {
		txinfo->pad[0] = sta_info->stats[tid].ampdu_len;
		txinfo->pad[1] = sta_info->stats[tid].ampdu_ack_len;
		txinfo->flags |= IEEE80211_TX_STAT_AMPDU;
		sta_info->stats[tid].clear = true;
	}

	rcu_read_unlock();

out_rcu:
	rcu_read_unlock();
}

void ar9170_tx_status(struct ar9170 *ar, struct sk_buff *skb,
		      const bool success)
{
	struct ieee80211_tx_info *txinfo;
	struct ar9170_tx_info *arinfo;

	ar9170_tx_accounting_free(ar, skb);

	txinfo = IEEE80211_SKB_CB(skb);
	arinfo = (void *) txinfo->rate_driver_data;

	if (success)
		txinfo->flags |= IEEE80211_TX_STAT_ACK;

	if (txinfo->flags & IEEE80211_TX_CTL_AMPDU)
		ar9170_tx_status_process_ampdu(ar, skb, txinfo);

	kref_put(&arinfo->ref, ar9170_tx_release);
}

void ar9170_tx_callback(struct ar9170 *ar, struct sk_buff *skb)
{
	struct ieee80211_tx_info *info = IEEE80211_SKB_CB(skb);
	struct ar9170_tx_info *arinfo = (void *) info->rate_driver_data;

	atomic_dec(&ar->tx_total_pending);

	if (info->flags & IEEE80211_TX_CTL_AMPDU)
		atomic_dec(&ar->tx_ampdu_pending);

	kref_put(&arinfo->ref, ar9170_tx_release);
}

static struct sk_buff *ar9170_get_queued_skb(struct ar9170 *ar,
					     const u8 cookie,
					     struct sk_buff_head *queue)
{
	unsigned long flags;
	struct sk_buff *skb;

	spin_lock_irqsave(&queue->lock, flags);
	skb_queue_walk(queue, skb) {
		struct _carl9170_tx_superframe *txc = (void *) skb->data;

		if (txc->s.cookie != cookie) {
#ifdef AR9170_QUEUE_DEBUG
			printk(KERN_DEBUG "%s: skip frame => cookie %x != %x\n",
			       wiphy_name(ar->hw->wiphy), cookie, txc->s.cookie);
			ar9170_print_txheader(ar, skb);
#endif /* AR9170_QUEUE_DEBUG */
			continue;
		}

		__skb_unlink(skb, queue);
		spin_unlock_irqrestore(&queue->lock, flags);
		return skb;
	}
	spin_unlock_irqrestore(&queue->lock, flags);

#ifdef AR9170_QUEUE_DEBUG
	printk(KERN_ERR "%s: cookie:%x is not in queue anymore.\n",
	       wiphy_name(ar->hw->wiphy), cookie);
	__ar9170_dump_txqueue(ar, queue);
#endif /* AR9170_QUEUE_DEBUG */

	return NULL;
}

static void ar9170_tx_fill_rateinfo(struct ar9170 *ar,
				    const unsigned int rix,
				    const unsigned int tries,
				    struct ieee80211_tx_info *txinfo)
{
	unsigned int i;

	for (i = 0; i < IEEE80211_TX_MAX_RATES; i++) {
		if (txinfo->status.rates[i].idx < 0)
			break;

		if (i == rix) {
			txinfo->status.rates[i].count = tries;
			i++;
			break;
		}
	}

	for (; i < IEEE80211_TX_MAX_RATES; i++) {
		txinfo->status.rates[i].idx = -1;
		txinfo->status.rates[i].count = 0;
	}
}

static void ar9170_bump_expired_frames(struct ar9170 *ar)
{
	struct carl9170_tx_stuck_cmd stuck;
	struct sk_buff_head *queue;
	struct sk_buff *skb;
	unsigned long flags;
	bool bump = false;
	uint8_t cookie;
	int err, i;

	for (i = 0; i < ar->hw->queues; i++) {
		queue = &ar->tx_status[i];

		spin_lock_irqsave(&queue->lock, flags);
		skb = skb_peek(queue);
		if (skb) {
			struct ieee80211_tx_info *txinfo;
			struct ar9170_tx_info *arinfo;

			txinfo = IEEE80211_SKB_CB(skb);
			arinfo = (void *) txinfo->rate_driver_data;

			if (time_before(arinfo->timeout +
			    msecs_to_jiffies(AR9170_BUMP_QUEUE), jiffies)) {
				struct _carl9170_tx_superframe *super;

				super = (void *) skb->data;
				cookie = super->s.cookie;
				bump = true;
			}
		}
		spin_unlock_irqrestore(&queue->lock, flags);

		if (!bump)
			continue;

		stuck.queue = ar9170_qmap[i];
		mutex_lock(&ar->mutex);
		err = ar->exec_cmd(ar, CARL9170_CMD_TX_STUCK, sizeof(stuck),
			   &stuck, sizeof(stuck), &stuck);
		mutex_unlock(&ar->mutex);
		if (err)
			ieee80211_queue_work(ar->hw, &ar->restart_work);
	}
}

static void ar9170_check_queue_stop_timeout(struct ar9170 *ar)
{
	unsigned long flags;
	int i;
	bool restart = false;

	spin_lock_irqsave(&ar->tx_stats_lock, flags);
	for (i = 0; i < ar->hw->queues; i++) {
		if (ieee80211_queue_stopped(ar->hw, i) &&
		    (ar->tx_stats[i].len >= AR9170_NUM_TX_LIMIT_SOFT) &&
		    (time_before(ar->queue_stop_timeout[i] +
		     msecs_to_jiffies(AR9170_QUEUE_STUCK_TIMEOUT), jiffies)))
			restart = true;
	}
	spin_unlock_irqrestore(&ar->tx_stats_lock, flags);

	if (restart)
		ieee80211_queue_work(ar->hw, &ar->restart_work);
}

/*
 * This worker tries to keeps an maintain tx_status queues.
 * So we can guarantee that incoming tx_status reports are
 * actually for a pending frame.
 */

void ar9170_tx_janitor(struct work_struct *work)
{
	struct ar9170 *ar = container_of(work, struct ar9170,
					 tx_janitor.work);
	if (unlikely(!IS_STARTED(ar)))
		return ;

	ar->tx_janitor_last_run = jiffies;

	ar9170_bump_expired_frames(ar);
	ar9170_check_queue_stop_timeout(ar);

	if (!atomic_read(&ar->tx_total_queued))
		return;

	ieee80211_queue_delayed_work(ar->hw,
				     &ar->tx_janitor,
				     msecs_to_jiffies(AR9170_JANITOR_DELAY));

	ar9170_tx_ampdu(ar);
	ar9170_tx(ar);
}

void ar9170_tx_process_status(struct ar9170 *ar,
			      const struct ar9170_rsp *cmd)
{	struct sk_buff *skb;
	struct ieee80211_tx_info *txinfo;
	struct ar9170_tx_info *arinfo;
	unsigned int q, i, j, r, t;
	bool success = true;

	j = 0;
	i = __le16_to_cpu(cmd->sequence);

	while (j < i) {
		if (unlikely(WARN_ON(j > ((cmd->len / 2) + 1)))) {
			print_hex_dump_bytes("UU:", DUMP_PREFIX_NONE,
					     (void *) cmd, cmd->len + 4);
			break;
		}

		q = ar9170_qmap[cmd->_tx_status[j].info &
			CARL9170_TX_STATUS_QUEUE_MASK];

		skb = ar9170_get_queued_skb(ar, cmd->_tx_status[j].cookie,
					    &ar->tx_status[q]);

		if (unlikely(WARN_ON_ONCE(!skb))) {
#ifdef AR9170_QUEUE_DEBUG
			printk(KERN_DEBUG "%s: orphaned cookie: %x q:%d\n",
			       wiphy_name(ar->hw->wiphy),
			       cmd->_tx_status[j].cookie, q);

			print_hex_dump_bytes("NF:", DUMP_PREFIX_NONE,
				(void *) cmd, cmd->len + 4);
#endif /* AR9170_QUEUE_DEBUG */

			j++;
			continue;
		}

		ar9170_release_dev_space(ar, skb);

#ifdef AR9170_QUEUE_DEBUG
		printk(KERN_DEBUG "%s: recv tx_status[%d] => %x, r:%d\n",
		       wiphy_name(ar->hw->wiphy), j, cmd->_tx_status[j].cookie,
				  cmd->_tx_status[j].info);
		ar9170_print_txheader(ar, skb);
#endif /* AR9170_QUEUE_DEBUG */

		txinfo = IEEE80211_SKB_CB(skb);
		arinfo = (void *) txinfo->rate_driver_data;

		if (!(cmd->_tx_status[j].info & CARL9170_TX_STATUS_SUCCESS))
			success = false;

		r = (cmd->_tx_status[j].info & CARL9170_TX_STATUS_RIX_MASK) >>
		    CARL9170_TX_STATUS_RIX_SHIFT;

		t = (cmd->_tx_status[j].info & CARL9170_TX_STATUS_TRIES_MASK) >>
		    CARL9170_TX_STATUS_TRIES_SHIFT;

		if (likely(t != 0)) {
			ar9170_tx_fill_rateinfo(ar, r, t, txinfo);
		} else {
			if (ar9170_nag_limiter(ar)) {
				printk(KERN_DEBUG "%s: firmware messed up "
				       "rate control fields (idx:%d).\n",
				       wiphy_name(ar->hw->wiphy), j);
			}
		}

		ar9170_tx_status(ar, skb, success);
		j++;
	}

	ar9170_tx_ampdu(ar);
	ar9170_tx(ar);
}

static __le32 ar9170_tx_physet(struct ar9170 *ar,
			       struct ieee80211_tx_info *info,
			       struct ieee80211_tx_rate *txrate)
{
	struct ieee80211_rate *rate = NULL;
	u32 power, chains;
	__le32 tmp;

	tmp = cpu_to_le32(0);

	if (txrate->flags & IEEE80211_TX_RC_40_MHZ_WIDTH)
		tmp |= cpu_to_le32(AR9170_TX_PHY_BW_40MHZ <<
			AR9170_TX_PHY_BW_SHIFT);
	/* this works because 40 MHz is 2 and dup is 3 */
	if (txrate->flags & IEEE80211_TX_RC_DUP_DATA)
		tmp |= cpu_to_le32(AR9170_TX_PHY_BW_40MHZ_DUP <<
			AR9170_TX_PHY_BW_SHIFT);

	if (txrate->flags & IEEE80211_TX_RC_SHORT_GI)
		tmp |= cpu_to_le32(AR9170_TX_PHY_SHORT_GI);

	if (txrate->flags & IEEE80211_TX_RC_MCS) {
		u32 r = txrate->idx;
		u8 *txpower;

		if (!(info->flags & IEEE80211_TX_CTL_AMPDU))
			WARN_ON_ONCE(1);

		/* heavy clip control */
		tmp |= cpu_to_le32((r & 0x7) <<
			AR9170_TX_PHY_TX_HEAVY_CLIP_SHIFT);

		if (txrate->flags & IEEE80211_TX_RC_40_MHZ_WIDTH) {
			if (info->band == IEEE80211_BAND_5GHZ)
				txpower = ar->power_5G_ht40;
			else
				txpower = ar->power_2G_ht40;
		} else {
			if (info->band == IEEE80211_BAND_5GHZ)
				txpower = ar->power_5G_ht20;
			else
				txpower = ar->power_2G_ht20;
		}

		power = txpower[r & 7];

		/* +1 dBm for HT40 */
		if (txrate->flags & IEEE80211_TX_RC_40_MHZ_WIDTH)
			power += 2;

		r <<= AR9170_TX_PHY_MCS_SHIFT;
		BUG_ON(r & ~AR9170_TX_PHY_MCS_MASK);

		tmp |= cpu_to_le32(r & AR9170_TX_PHY_MCS_MASK);
		tmp |= cpu_to_le32(AR9170_TX_PHY_MOD_HT);

		if (txrate->flags & IEEE80211_TX_RC_GREEN_FIELD)
			tmp |= cpu_to_le32(AR9170_TX_PHY_GREENFIELD);
	} else {
		u8 *txpower;
		u32 mod;
		u32 phyrate;
		u8 idx = txrate->idx;

		if (info->band != IEEE80211_BAND_2GHZ) {
			idx += 4;
			txpower = ar->power_5G_leg;
			mod = AR9170_TX_PHY_MOD_OFDM;
		} else {
			if (idx < 4) {
				txpower = ar->power_2G_cck;
				mod = AR9170_TX_PHY_MOD_CCK;
			} else {
				mod = AR9170_TX_PHY_MOD_OFDM;
				txpower = ar->power_2G_ofdm;
			}
		}

		rate = &__ar9170_ratetable[idx];

		phyrate = rate->hw_value & 0xF;
		power = txpower[(rate->hw_value & 0x30) >> 4];
		phyrate <<= AR9170_TX_PHY_MCS_SHIFT;

		tmp |= cpu_to_le32(mod);
		tmp |= cpu_to_le32(phyrate);

		/*
		 * short preamble seems to be broken
		 *
		 * if (txrate->flags & IEEE80211_TX_RC_USE_SHORT_PREAMBLE)
		 *	tmp |= cpu_to_le32(AR9170_TX_PHY_SHORT_PREAMBLE);
		 */
	}


	power <<= AR9170_TX_PHY_TX_PWR_SHIFT;
	power &= AR9170_TX_PHY_TX_PWR_MASK;
	tmp |= cpu_to_le32(power);

	/* set TX chains */
	if (ar->eeprom.tx_mask == 1) {
		chains = AR9170_TX_PHY_TXCHAIN_1;
	} else {
		chains = AR9170_TX_PHY_TXCHAIN_2;

		/* >= 36M legacy OFDM - use only one chain */
		if (rate && rate->bitrate >= 360 &&
		    !(txrate->flags & IEEE80211_TX_RC_MCS))
			chains = AR9170_TX_PHY_TXCHAIN_1;
	}
	tmp |= cpu_to_le32(chains << AR9170_TX_PHY_TXCHAIN_SHIFT);

	return tmp;
}

static int ar9170_tx_prepare(struct ar9170 *ar, struct sk_buff *skb)
{
	struct ieee80211_hdr *hdr;
	struct _carl9170_tx_superframe *txc;
	struct ieee80211_tx_info *info;
	struct ieee80211_tx_rate *txrate;
	struct ar9170_tx_info *arinfo;
	unsigned int hw_queue;
	int i;
	u16 keytype = 0;
	u16 len, icv = 0;

	BUILD_BUG_ON(sizeof(*arinfo) > sizeof(info->rate_driver_data));
	BUILD_BUG_ON(sizeof(struct _carl9170_tx_superdesc) !=
		     AR9170_TX_SUPERDESC_LEN);

	BUILD_BUG_ON(sizeof(struct _ar9170_tx_hwdesc) !=
		     AR9170_TX_HWDESC_LEN);

	hw_queue = ar9170_get_queue(ar, skb);

	hdr = (void *)skb->data;
	info = IEEE80211_SKB_CB(skb);
	len = skb->len;

	txc = (void *)skb_push(skb, sizeof(*txc));
	memset(txc, 0, sizeof(*txc));

	if (info->control.hw_key) {
		icv = info->control.hw_key->icv_len;

		switch (info->control.hw_key->alg) {
		case ALG_WEP:
			keytype = AR9170_TX_MAC_ENCR_RC4;
			break;
		case ALG_TKIP:
			keytype = AR9170_TX_MAC_ENCR_RC4;
			break;
		case ALG_CCMP:
			keytype = AR9170_TX_MAC_ENCR_AES;
			break;
		default:
			WARN_ON(1);
			goto err_out;
		}
	}

	/* Length */
	txc->s.len = cpu_to_le16(len + sizeof(*txc));
	txc->f.length = cpu_to_le16(len + icv + 4);

	txc->f.mac_control = cpu_to_le16(AR9170_TX_MAC_HW_DURATION |
					 AR9170_TX_MAC_BACKOFF);
	txc->f.mac_control |= cpu_to_le16(ar9170_qmap[hw_queue] <<
		AR9170_TX_MAC_QOS_SHIFT);

	txc->f.mac_control |= cpu_to_le16(keytype);
	txc->f.phy_control = cpu_to_le32(0);

	if (info->flags & IEEE80211_TX_CTL_NO_ACK)
		txc->f.mac_control |= cpu_to_le16(AR9170_TX_MAC_NO_ACK);

	if (info->flags & IEEE80211_TX_CTL_SEND_AFTER_DTIM)
		txc->f.mac_control |= cpu_to_le16(AR9170_TX_MAC_CAB);

	txrate = &info->control.rates[0];
	if (txrate->flags & IEEE80211_TX_RC_USE_RTS_CTS)
		txc->f.mac_control |= cpu_to_le16(AR9170_TX_MAC_PROT_RTS);
	else if (txrate->flags & IEEE80211_TX_RC_USE_CTS_PROTECT)
		txc->f.mac_control |= cpu_to_le16(AR9170_TX_MAC_PROT_CTS);

	txc->s.tries[0] = txrate->count;
	txc->f.phy_control |= ar9170_tx_physet(ar, info, txrate);

	for (i = 1; i < IEEE80211_TX_MAX_RATES; i++) {
		txrate = &info->control.rates[i];
		if (txrate->idx < 0)
			break;

		if (i > CARL9170_TX_MAX_RATES)
			break;

		txc->s.tries[i] = txrate->count;
		txc->s.rr[i - 1] = ar9170_tx_physet(ar, info, txrate);
	}

	if (info->flags & IEEE80211_TX_CTL_AMPDU) {
		unsigned int density, factor;

		if (unlikely(!info->control.sta))
			goto err_out;

		density = info->control.sta->ht_cap.ampdu_density;
		factor = info->control.sta->ht_cap.ampdu_factor;

		if (density) {
			/*
			 * Watch out!
			 *
			 * Otus uses slightly different density values than
			 * those from the 802.11n spec.
			 */

			density = max_t(unsigned int, density + 1, 7);
		}

		factor = min_t(unsigned int, 1, factor);

		SET_VAL(AR9170_TX_SUPER_AMPDU_DENSITY,
			txc->s.ampdu_settings, density);

		SET_VAL(AR9170_TX_SUPER_AMPDU_FACTOR,
			txc->s.ampdu_settings, factor);

		if (info->control.rates[0].flags & IEEE80211_TX_RC_MCS)
			txc->f.mac_control |= cpu_to_le16(AR9170_TX_MAC_AGGR);
		else
			WARN_ON_ONCE(1);
	}

	arinfo = (void *)info->rate_driver_data;
	arinfo->timeout = jiffies;
	arinfo->ar = ar;
	kref_init(&arinfo->ref);
	return 0;

err_out:
	skb_pull(skb, sizeof(*txc));
	return -EINVAL;
}

static void ar9170_tx_indicate_immba(struct ar9170 *ar, struct sk_buff *skb,
				     bool set)
{
	struct _carl9170_tx_superframe *super;

	if (unlikely(!skb))
		return ;

	super = (void *) skb->data;

	if (set)
		super->f.mac_control |= cpu_to_le16(AR9170_TX_MAC_IMM_BA);
	else
		super->f.mac_control &= cpu_to_le16(~AR9170_TX_MAC_IMM_BA);

}

static void ar9170_tx_indicate_ampdu(struct ar9170 *ar, struct sk_buff *skb,
				     bool factor, bool density)
{
	struct _carl9170_tx_superframe *super;

	if (unlikely(!skb))
		return ;

	super = (void *) skb->data;

	if (factor)
		super->s.ampdu_settings |= AR9170_TX_SUPER_AMPDU_COMMIT_FACTOR;

	if (density)
		super->s.ampdu_settings |= AR9170_TX_SUPER_AMPDU_COMMIT_DENSITY;

}

static void ar9170_tx_ampdu(struct ar9170 *ar)
{
	struct sk_buff_head agg;
	struct ar9170_sta_tid *tid_info;
	struct sk_buff *skb, *first;
	unsigned int i;
	unsigned long flags;
	u16 seq, queue, tmpssn;

	if (atomic_read(&ar->tx_ampdu_pending)) {
#ifdef AR9170_TXAGG_DEBUG
		printk(KERN_DEBUG "%s: aggregation in progress. %d\n",
		       wiphy_name(ar->hw->wiphy),
		       atomic_read(&ar->tx_ampdu_pending));
#endif /* AR9170_TXAGG_DEBUG */
		return ;
	}

	__skb_queue_head_init(&agg);

	i = 0;
	rcu_read_lock();
	list_for_each_entry_rcu(tid_info, &ar->tx_ampdu_list, list) {
		if (unlikely(i++ > 128))
			break;

		queue = TID_TO_WME_AC(tid_info->tid);
		spin_lock_irqsave(&tid_info->lock, flags);
		if (tid_info->state < AR9170_TID_STATE_PROGRESS) {
#ifdef AR9170_TXAGG_DEBUG
			printk(KERN_DEBUG "%s: dangling aggregation entry!\n",
			       wiphy_name(ar->hw->wiphy));
#endif /* AR9170_TXAGG_DEBUG */
			goto processed;
		}

		if (tid_info->state != AR9170_TID_STATE_XMIT) {
			first = skb_peek(&tid_info->queue);

			if (first) {
				struct ieee80211_tx_info *txinfo;
				struct ar9170_tx_info *arinfo;

				txinfo = IEEE80211_SKB_CB(first);
				arinfo = (void *) txinfo->rate_driver_data;

				if (time_after(arinfo->timeout +
				    msecs_to_jiffies(AR9170_QUEUE_TIMEOUT),
				    jiffies))
					goto processed;

				tid_info->snx = ar9170_get_seq(first);
				tid_info->state = AR9170_TID_STATE_XMIT;
			} else {
				goto processed;
			}
		}

		seq = ar->tx_ampdu_list_len;
		if (!seq || !tid_info->counter++ % seq)
			goto processed;

		first = skb_peek(&tid_info->queue);
		tmpssn = ar9170_get_seq(first);
		seq = tid_info->snx;

		if (unlikely(tmpssn != seq)) {
#ifdef AR9170_TXAGG_DEBUG
			printk(KERN_DEBUG "%s: ssn mismatch [%d != %d]\n.",
			       wiphy_name(ar->hw->wiphy), seq, tmpssn);
#endif /* AR9170_TXAGG_DEBUG */
			tid_info->state = AR9170_TID_STATE_IDLE;

			goto processed;
		}

#ifdef AR9170_TXAGG_DEBUG
		printk(KERN_DEBUG "%s: generate A-MPDU for tid:%d ssn:%d with "
		       "%d queued frames.\n", wiphy_name(ar->hw->wiphy),
		       tid_info->tid, tid_info->snx,
		       skb_queue_len(&tid_info->queue));
		__ar9170_dump_txqueue(ar, &tid_info->queue);
#endif /* AR9170_TXAGG_DEBUG */

		while ((skb = skb_peek(&tid_info->queue))) {
			/* strict seq 0, 1, ..., n - 1, n frame seq order */
			if (unlikely(ar9170_get_seq(skb) != seq))
				break;

			/* don't upload more than AMPDU FACTOR allows */
			if (unlikely(SEQ_DIFF(tid_info->snx, tid_info->bsn) >=
			    (tid_info->max - 1)))
				break;

			atomic_inc(&ar->tx_ampdu_pending);
			tid_info->snx = seq = SEQ_NEXT(seq);
			__skb_unlink(skb, &tid_info->queue);

			__skb_queue_tail(&agg, skb);

			if (skb_queue_len(&agg) >= AR9170_NUM_TX_AGG_MAX)
				break;
		}

		if (skb_queue_empty(&tid_info->queue) ||
		    ar9170_get_seq(skb_peek(&tid_info->queue)) != tid_info->snx) {
			/*
			 * deactive the TID, whenever the queue is empty,
			 * or we are still missing some of the A-MPDU frames
			 */

			tid_info->state = AR9170_TID_STATE_IDLE;
		}

processed:
		spin_unlock_irqrestore(&tid_info->lock, flags);

		if (tid_info->counter % 4 == 1) {
			/* apply ampdu settings */
			ar9170_tx_indicate_ampdu(ar, skb_peek(&agg), true, true);
		}

		/* set aggregation push bit */
		ar9170_tx_indicate_immba(ar, skb_peek_tail(&agg), true);

#ifdef AR9170_TXAGG_DEBUG
		printk(KERN_DEBUG "%s: generated A-MPDU looks like this:\n",
			wiphy_name(ar->hw->wiphy));
		__ar9170_dump_txqueue(ar, &agg);
#endif /* AR9170_TXAGG_DEBUG */

		spin_lock_irqsave(&ar->tx_pending[queue].lock, flags);
		skb_queue_splice_tail_init(&agg, &ar->tx_pending[queue]);
		spin_unlock_irqrestore(&ar->tx_pending[queue].lock, flags);
	}

	rcu_read_unlock();
}

static struct sk_buff *ar9170_tx_pick_skb(struct ar9170 *ar,
					  struct sk_buff_head *queue)
{
	unsigned long flags;
	struct sk_buff *skb;
	struct ieee80211_tx_info *info;
	struct ar9170_tx_info *arinfo;
	struct _carl9170_tx_superframe *super;
	int cookie;

	BUILD_BUG_ON(sizeof(*arinfo) > sizeof(info->rate_driver_data));

	spin_lock_irqsave(&queue->lock, flags);
	skb = skb_peek(queue);
	if (unlikely(!skb))
		goto err_unlock;

	cookie = ar9170_alloc_dev_space(ar, skb->len);
	if (unlikely(cookie < 0)) {
		int i;

		/* meh, no more cookies >-/ */
		ieee80211_stop_queues(ar->hw);
		for (i = 0; i < ar->hw->queues; i++)
			ar->queue_stop_timeout[i] = jiffies;

		goto err_unlock;
	}
	__skb_unlink(skb, queue);
	spin_unlock_irqrestore(&queue->lock, flags);

	super = (void *) skb->data;
	info = IEEE80211_SKB_CB(skb);
	arinfo = (void *) info->rate_driver_data;

	arinfo->timeout = jiffies;

	/* Yummy... Cookie! */
	super->s.cookie = (u8) cookie;

	/*
	 * increase ref count to "2".
	 * Ref counting is the easiest way to solve the race between
	 * the the urb's completion routine: ar9170_tx_callback and
	 * wlan's status functions: ar9170_txrate_status/janitor.
	 */
	kref_get(&arinfo->ref);

	return skb;

err_unlock:
	spin_unlock_irqrestore(&queue->lock, flags);
	return NULL;
}

static void ar9170_tx(struct ar9170 *ar)
{
	struct sk_buff *skb;
	unsigned int i, q;
	int err;
	bool schedule_garbagecollector = false;

	if (unlikely(!IS_STARTED(ar)))
		return ;

	for (i = 0; i < ar->hw->queues; i++) {
		while (!skb_queue_empty(&ar->tx_pending[i])) {

			skb = ar9170_tx_pick_skb(ar, &ar->tx_pending[i]);
			if (unlikely(!skb))
				break;

			q = __ar9170_get_queue(ar, i);
			atomic_inc(&ar->tx_total_pending);

#ifdef AR9170_QUEUE_DEBUG
			printk(KERN_DEBUG "%s: send frame q:%d =>\n",
			       wiphy_name(ar->hw->wiphy), q);
			ar9170_print_txheader(ar, skb);
#endif /* AR9170_QUEUE_DEBUG */

			/*
			 * NB: tx_status[i] vs. tx_status[q]
			 */
			skb_queue_tail(&ar->tx_status[q], skb);

			err = ar->tx(ar, skb);
			if (unlikely(err)) {
				skb_unlink(skb, &ar->tx_status[q]);
				ar9170_tx_callback(ar, skb);
				ar9170_release_dev_space(ar, skb);
				ar9170_tx_status(ar, skb, false);
			} else {
				schedule_garbagecollector = true;
			}
		}

#ifdef AR9170_QUEUE_DEBUG
		printk(KERN_DEBUG "%s: ar9170_tx report for queue %d\n",
		       wiphy_name(ar->hw->wiphy), i);

		printk(KERN_DEBUG "%s: unprocessed pending frames left:\n",
		       wiphy_name(ar->hw->wiphy));
		ar9170_dump_txqueue(ar, &ar->tx_pending[i]);
#endif /* AR9170_QUEUE_DEBUG */
	}

	if (!schedule_garbagecollector)
		return;

	ieee80211_queue_delayed_work(ar->hw,
				     &ar->tx_janitor,
				     msecs_to_jiffies(AR9170_JANITOR_DELAY));
}

static bool ar9170_tx_ampdu_queue(struct ar9170 *ar, struct ieee80211_sta *sta,
				  struct sk_buff *skb)
{
	struct ar9170_sta_info *sta_info;
	struct ar9170_sta_tid *agg;
	struct sk_buff *iter;
	unsigned long flags;
	unsigned int max;
	u16 tid, seq, qseq, off;
	bool run = false;

	tid = ar9170_get_tid(skb);
	seq = ar9170_get_seq(skb);
	sta_info = (void *) sta->drv_priv;

	rcu_read_lock();
	agg = rcu_dereference(sta_info->agg[tid]);
	max = sta_info->ampdu_max_len;

	if (!agg) {
#ifdef AR9170_TXAGG_DEBUG
		printk(KERN_DEBUG "%s: BlockACK session not available "
		       "for ESS:%pM tid:%d state:%d.\n",
		       wiphy_name(ar->hw->wiphy), sta->addr, tid, agg->state);
#endif /* AR9170_TXAGG_DEBUG */
		goto err_unlock_rcu;
	}

	spin_lock_irqsave(&agg->lock, flags);
	if (unlikely(agg->state < AR9170_TID_STATE_IDLE)) {
#ifdef AR9170_TXAGG_DEBUG
		printk(KERN_DEBUG "%s: BlockACK session not fully initialized "
		       "for ESS:%pM tid:%d state:%d.\n",
		       wiphy_name(ar->hw->wiphy), sta->addr, tid, agg->state);
#endif /* AR9170_TXAGG_DEBUG */
		goto err_unlock;
	}

	/* check if seq is within the BA window */
	if (unlikely(!BAW_WITHIN(agg->bsn, AR9170_BAW_BITS, seq))) {
#ifdef AR9170_TXAGG_DEBUG
		printk(KERN_DEBUG "%s: BAS:[%pM,t:%d] frame with "
		       "seq:%d does not fit into BA window (%d - %d)\n",
		       wiphy_name(ar->hw->wiphy), sta->addr,
		       tid, seq, agg->bsn, (agg->snx + max) & 0x0fff);
#endif /* AR9170_TXAGG_DEBUG */
		goto err_unlock;
	}

	if (unlikely(!BAW_WITHIN(agg->snx, AR9170_BAW_BITS, seq))) {
#ifdef AR9170_TXAGG_DEBUG
		printk(KERN_DEBUG "%s: BAS:[%pM,t:%d] frame with "
		       "seq:%d arrived too late.\n",
		       wiphy_name(ar->hw->wiphy), sta->addr,
		       tid, seq);
#endif /* AR9170_TXAGG_DEBUG */
		goto err_unlock;
	}

	off = SEQ_DIFF(seq, agg->bsn);
	if (test_and_set_bit(off, agg->bitmap)) {
#ifdef AR9170_TXAGG_DEBUG
		printk(KERN_DEBUG "%s: BAS:[%pM,t:%d] frame[s:%d]'s "
		       "BAW offset:%d is already used.\n",
		       wiphy_name(ar->hw->wiphy), sta->addr,
		       agg->tid, seq, off);
#endif /* AR9170_TXAGG_DEBUG */
		goto err_unlock;
	}

	if (likely(BAW_WITHIN(agg->hsn, AR9170_BAW_BITS, seq))) {
		__skb_queue_tail(&agg->queue, skb);
		agg->hsn = seq;
		goto queued;
	}

	skb_queue_reverse_walk(&agg->queue, iter) {
		qseq = ar9170_get_seq(iter);

		if (BAW_WITHIN(qseq, AR9170_BAW_BITS, seq)) {
			__skb_queue_after(&agg->queue, iter, skb);
			goto queued;
		}
	}

	__skb_queue_head(&agg->queue, skb);
queued:

	if (unlikely(agg->state != AR9170_TID_STATE_XMIT)) {
		if (agg->snx == ar9170_get_seq(skb_peek(&agg->queue))) {
			agg->state = AR9170_TID_STATE_XMIT;
			run = true;
		} else {
#ifdef AR9170_TXAGG_DEBUG
			printk(KERN_DEBUG "%s: BAS:[%pM,t:%d] PreAGG reorder, "
			       "awaiting:%d, current head:%d, queued:%d\n",
			       wiphy_name(ar->hw->wiphy), sta->addr, agg->tid,
			       agg->snx, ar9170_get_seq(skb_peek(&agg->queue)),
			       seq);
#endif /* AR9170_TXAGG_DEBUG */
		}
	}
	spin_unlock_irqrestore(&agg->lock, flags);
	rcu_read_unlock();

	return run;

err_unlock:
	spin_unlock_irqrestore(&agg->lock, flags);

err_unlock_rcu:
	rcu_read_unlock();

	ar9170_tx_status(ar, skb, false);
	return false;
}

int ar9170_op_tx(struct ieee80211_hw *hw, struct sk_buff *skb)
{
	struct ar9170 *ar = hw->priv;
	struct ieee80211_tx_info *info;
	struct ieee80211_sta *sta;
	bool run;

	if (unlikely(!IS_STARTED(ar)))
		goto err_free;

	info = IEEE80211_SKB_CB(skb);
	sta = info->control.sta;

	if (unlikely(ar9170_tx_prepare(ar, skb)))
		goto err_free;

	ar9170_tx_accounting(ar, skb);
	/*
	 * from now on, one has to use ar9170_tx_status to free
	 * all ressouces which are associated with the frame.
	 */

	if (info->flags & IEEE80211_TX_CTL_AMPDU) {
		run = ar9170_tx_ampdu_queue(ar, sta, skb);

		if (run)
			ar9170_tx_ampdu(ar);
	} else {
		unsigned int queue = skb_get_queue_mapping(skb);

		skb_queue_tail(&ar->tx_pending[queue], skb);
	}

	ar9170_tx(ar);

	return NETDEV_TX_OK;

err_free:
	dev_kfree_skb_any(skb);
	return NETDEV_TX_OK;
}
