/*
 * Firmware Header definitions
 *
 * Copyright 2009, Christian Lamparter <chunkeey@googlemail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, see
 * http://www.gnu.org/licenses/.
 */

#ifndef __SHARED_FWHDR_H
#define __SHARED_FWHDR_H

/* NOTE: Don't mess with the order of the flags! */
enum carl9170_fw_feature_list_v1 {
	/* Always set */
	CARL9170_FW_DUMMY_FEATURE,

	/* Firmware descriptor header is *integrated* into firmware image */
	CARL9170_FW_INTEGRATED_DESC_HEADER,

	/* CRC32 checksum at the end */
	CARL9170_FW_CRC32_CHECKSUM_TAIL,

	/* unusable - reserved to flag non-functional debug firmwares */
	CARL9170_FW_UNUSEABLE,

	/*
	 * AR9170_CMD_ECHO, AR9170_CMD_REBOOT, AR9170_CMD_SWRST
	 * AR9170_CMD_RREG, AR9170_CMD_WREG
	 */
	AR9170_FW_COMMAND_BASIC,

	/* AR9170_CMD_RF_INIT, AR9170_CMD_FREQ_START, AR9170_CMD_FREQUENCY */
	AR9170_FW_COMMAND_PHY,

	/* AR9170_CMD_EKEY, AR9170_CMD_DKEY */
	AR9170_FW_COMMAND_CAM,

	/* Firmware has a software Content After Beacon Queueing mechanism */
	CARL9170_FW_WLANTX_CAB,

	/* The firmware is capable of responding to incoming BAR frames */
	AR9170_FW_HANDLE_BACK_REQ,

	/* The firmware reorders incoming frames */
	AR9170_FW_WLANTX_DELAYED_TX,

	/* General API Compatibility */

	/*
	 * Legacy firmware
	 * This flag is set to indicate that this firmware understands the
	 * legacy tx descriptors and would probably work with the original
	 * OTUS driver.
	 */
	AR9170_FW_LEGACY,

	/*
	 * This firmware understands the ar9170_tx_superdesc header and
	 * provides more advanced xfer feedback reporting.
	 */
	CARL9170_FW_CARL9170_V1,

	/* KEEP LAST */
	__CARL9170_FW_FEATURE_NUM
};

enum carl9170_fw_usb_feature_list_v1 {
	/* Always set */
	CARL9170_FW_USB_DUMMY_FEATURE,

	/*
	 * Indicates that this image has special boot block which prevents
	 * legacy drivers to drive the firmware.
	 */
	CARL9170_FW_USB_MINIBOOT,

	/* usb registers are initialized by the firmware */
	AR9170_FW_USB_INIT_FIRMWARE,

	/* command traps & notifications are send through EP2 */
	AR9170_FW_USB_RESP_EP2,

	/* usb download (app -> fw) stream */
	AR9170_FW_USB_DOWN_STREAM,

	/* usb upload (fw -> app) stream */
	AR9170_FW_USB_UP_STREAM,

	/* KEEP LAST */
	__CARL9170_FW_USB_FEATURE_NUM
};

#define OTUS_MAGIC	"OTAR"
#define USB_MAGIC	"USB\0"
#define MOTD_MAGIC	"MOTD"
#define FIX_MAGIC	"FIX\0"
#define DBG_MAGIC	"DBG\0"
#define LAST_MAGIC	"LAST"

#define CARL9170_FW_SET_DAY(d) (((d) - 1) % 31)
#define CARL9170_FW_SET_MONTH(m) ((((m) - 1) % 12) * 31)
#define CARL9170_FW_SET_YEAR(y) (((y) - 10) * 372)

#define CARL9170_FW_GET_DAY(d) (((d) % 31) + 1)
#define CARL9170_FW_GET_MONTH(m) ((((m) / 31) % 12) + 1)
#define CARL9170_FW_GET_YEAR(y) ((y) / 372 + 10)

struct carl9170_fw_desc_head {
	uint8_t	magic[4];
	__le16 length;
	uint8_t min_ver;
	uint8_t cur_ver;
} __packed;

#define CARL9170_FW_OTUS_DESC_MIN_VER		1
#define CARL9170_FW_OTUS_DESC_CUR_VER		1
struct carl9170_fw_otus_desc_v1 {
	struct carl9170_fw_desc_head head;
	__le32 fw_feature_set;
	__le32 bcn_addr;
	__le16 bcn_len;
	__le16 fw_year_month_day;
} __packed;

#define CARL9170_FW_USB_DESC_MIN_VER		1
#define CARL9170_FW_USB_DESC_CUR_VER		1
struct carl9170_fw_usb_desc_v1 {
	struct carl9170_fw_desc_head head;
	__le32 usb_feature_set;
	__le16 tx_frag_len;
	__le16 rx_max_frame_len;
	__le16 miniboot_size;
	uint8_t tx_descs;
	uint8_t free;
} __packed;

#define CARL9170_FW_MOTD_STRING_LEN			24
#define CARL9170_FW_MOTD_RELEASE_LEN			20
#define CARL9170_FW_MOTD_DESC_MIN_VER			1
#define CARL9170_FW_MOTD_DESC_CUR_VER			1
struct carl9170_fw_motd_desc_v1 {
	struct carl9170_fw_desc_head head;

	char desc[CARL9170_FW_MOTD_STRING_LEN];
	char release[CARL9170_FW_MOTD_RELEASE_LEN];
} __packed;

#define CARL9170_FW_FIX_DESC_MIN_VER			1
#define CARL9170_FW_FIX_DESC_CUR_VER			1
struct carl9170_fw_fix_entry {
	__le32 address;
	__le32 mask;
	__le32 value;
} __packed;

struct carl9170_fw_fix_desc_v1 {
	struct carl9170_fw_desc_head head;
	struct carl9170_fw_fix_entry data[0];
} __packed;

#define CARL9170_FW_DEBUG_DESC_MIN_VER			1
#define CARL9170_FW_DEBUG_DESC_CUR_VER			1
struct carl9170_fw_debug_desc_v1 {
	struct carl9170_fw_desc_head head;

	/* Put your debugging definitions here */
} __packed;

#define CARL9170_FW_LAST_DESC_MIN_VER			1
#define CARL9170_FW_LAST_DESC_CUR_VER			1
struct carl9170_fw_last_desc_v1 {
	struct carl9170_fw_desc_head head;
} __packed;

#define CARL9170_FW_DESC_MAX_LENGTH			8192

#define FILL_DESC(_magic, _length, _min_ver, _cur_ver)			\
	.head = {							\
		.magic = _magic,					\
		.length = cpu_to_le16(_length),				\
		.min_ver = _min_ver,					\
		.cur_ver = _cur_ver,					\
	}

#define for_each_hdr(desc, fw_desc)					\
	for (desc = fw_desc;						\
	     (memcmp(desc->magic, LAST_MAGIC, 4) && desc->length);	\
	     desc = (void *)((unsigned long)desc + desc->length))

#define CHECK_HDR_VERSION(head, _min_ver)				\
	(((head)->cur_ver < _min_ver) || ((head)->min_ver > _min_ver))	\

static inline bool is_feature_supported(__le32 list, uint8_t feature)
{
	return le32_to_cpu(list) & BIT(feature);
}

#endif /* __FWHDR_H */
