/*
 * carl9170 firmware - used by the ar9170 wireless device
 *
 * WLAN header definition
 *
 * Copyright (c) 2000-2005 ZyDAS Technology Corporation
 * Copyright (c) 2007-2009 Atheros Communications, Inc.
 * Copyright	2009	Johannes Berg <johannes@sipsolutions.net>
 * Copyright	2009	Christian Lamparter <chunkeey@googlemail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef __WLANTX_H
#define __WLANTX_H

#include "carl9170.h"

static inline void set_wlan_txq_dma_addr(const unsigned int q, const uint32_t v)
{
	set(AR9170_MAC_REG_DMA_TXQ_ADDR + (q << 3), v);
}

static inline void set_wlan_txq_dma_curr_addr(const unsigned int q, const uint32_t v)
{
	set(AR9170_MAC_REG_DMA_TXQ_CURR_ADDR + (q << 3), v);
}

static inline struct dma_desc *get_wlan_txq_dma_addr(const unsigned int q)
{
	return (void *)(get(AR9170_MAC_REG_DMA_TXQ_ADDR + (q << 3)));
}

static inline struct dma_desc *get_wlan_txq_addr(const unsigned int q)
{
	return (void *)(get(AR9170_MAC_REG_DMA_TXQ_CURR_ADDR + (q << 3)));
}

static inline void _wlan_trigger(const uint32_t queue_bit)
{
	set(AR9170_MAC_REG_DMA_TRIGGER, queue_bit);
}

#ifdef CONFIG_CARL9170FW_DELAYED_TX
static inline void wlan_trigger(const uint32_t queue_bit)
{
	fw.wlan.tx_trigger |= queue_bit;
}
#else
static inline void wlan_trigger(const uint32_t queue_bit)
{
	_wlan_trigger(queue_bit);
}
#endif /* CONFIG_CARL9170FW_DELAYED_TX */

static inline uint8_t ar9170_get_rx_macstatus_status(struct dma_desc *desc)
{
	return *((uint8_t *) DESC_PAYLOAD_OFF(desc->lastAddr,
		desc->lastAddr->dataSize - 1));
}

static inline uint8_t ar9170_get_rx_macstatus_error(struct dma_desc *desc)
{
	unsigned int offset;

	if (desc->lastAddr->dataSize == 1) {
		while (desc->lastAddr != desc->nextAddr)
			desc = desc->nextAddr;

		offset = desc->dataSize - 1;
	} else {
		desc = desc->lastAddr;
		offset = desc->dataSize -
			(sizeof(struct ar9170_rx_macstatus) -
			 offsetof(struct ar9170_rx_macstatus, error));
	}

	return *((uint8_t *) DESC_PAYLOAD_OFF(desc, offset));
}

static inline struct ieee80211_hdr *ar9170_get_rx_i3e(struct dma_desc *desc)
{
	if (!((ar9170_get_rx_macstatus_status(desc) &
		AR9170_RX_STATUS_MPDU_MASK) & AR9170_RX_STATUS_MPDU_LAST)) {
		return (void *)(DESC_PAYLOAD_OFF(desc,
			offsetof(struct ar9170_rx_frame_head, i3e)));
	} else {
		return (void *)(DESC_PAYLOAD_OFF(desc,
			offsetof(struct ar9170_rx_frame_tail, i3e)));
	}
}

static inline struct ar9170_rx_head *ar9170_get_rx_head(struct dma_desc *desc)
{
	if (!((ar9170_get_rx_macstatus_status(desc) &
		AR9170_RX_STATUS_MPDU_MASK) & AR9170_RX_STATUS_MPDU_LAST)) {
		return (void *)((uint8_t *)DESC_PAYLOAD(desc) +
			offsetof(struct ar9170_rx_frame_head, phy_head));
	} else {
		return (void *) NULL;
	}
}

static inline uint32_t ar9170_rx_to_phy(struct dma_desc *rx)
{
	struct ar9170_tx_hw_phy_control phy;
	struct ar9170_rx_head *head;
	uint8_t mac_status;

	phy.set = 0;

	head = ar9170_get_rx_head(rx);
	if (!head)
		return le32_to_cpu(phy.set);

	mac_status = ar9170_get_rx_macstatus_status(rx);

	phy.modulation = mac_status & AR9170_RX_STATUS_MODULATION_MASK;
	phy.chains = AR9170_TX_PHY_TXCHAIN_1;

	switch (phy.modulation) {
	case AR9170_RX_STATUS_MODULATION_CCK:
		if (mac_status & AR9170_RX_STATUS_SHORT_PREAMBLE)
			phy.preamble = 1;

		switch (head->plcp[0]) {
		case AR9170_RX_PHY_RATE_CCK_2M:
			phy.mcs = AR9170_TX_PHY_RATE_CCK_2M;
			break;

		case AR9170_RX_PHY_RATE_CCK_5M:
			phy.mcs = AR9170_TX_PHY_RATE_CCK_5M;
			break;

		case AR9170_RX_PHY_RATE_CCK_11M:
			phy.mcs = AR9170_TX_PHY_RATE_CCK_11M;
			break;

		case AR9170_RX_PHY_RATE_CCK_1M:
		default:
			phy.mcs = AR9170_TX_PHY_RATE_CCK_1M;
			break;

		}
		break;

	case AR9170_RX_STATUS_MODULATION_DUPOFDM:
	case AR9170_RX_STATUS_MODULATION_OFDM:
		phy.mcs = head->plcp[0] & 0xf;
		break;

	case AR9170_RX_STATUS_MODULATION_HT:
		if (head->plcp[3] & 0x80)
			phy.bandwidth = 2;

		if (head->plcp[6] & 0x80)
			phy.short_gi = 1;

		/* TODO: Enable both chains for MCS > 7 */
		phy.mcs = head->plcp[6] & 0x7;
		break;
	}

	return le32_to_cpu(phy.set);
}

static inline unsigned int ar9170_get_rx_mpdu_len(struct dma_desc *desc)
{
	/*
	 * WARNING: you have to check the error bits in macstatus first!
	 */

	unsigned int mpdu_len = desc->totalLen;

	mpdu_len -= sizeof(struct ar9170_rx_macstatus);

	switch (ar9170_get_rx_macstatus_status(desc) & AR9170_RX_STATUS_MPDU_MASK) {
	case AR9170_RX_STATUS_MPDU_LAST:
		mpdu_len -= sizeof(struct ar9170_rx_phystatus);
		break;

	case AR9170_RX_STATUS_MPDU_SINGLE:
		mpdu_len -= sizeof(struct ar9170_rx_phystatus);

	case AR9170_RX_STATUS_MPDU_FIRST:
		mpdu_len -= sizeof(struct ar9170_rx_head);
		break;

	case AR9170_RX_STATUS_MPDU_MIDDLE:
	default:
		break;
	}

	return mpdu_len;
}

static inline bool ar9170_tx_length_check(const uint16_t len)
{
	return len > (sizeof(struct carl9170_tx_superframe) + 24 +
			 FCS_LEN);
}

static inline struct carl9170_tx_superframe *get_super(struct dma_desc *desc)
{
	return container_of(DESC_PAYLOAD(desc), struct carl9170_tx_superframe,
			    f);
}

static inline void hide_super(struct dma_desc *desc)
{
	desc->dataAddr = (uint8_t *)
		(((unsigned long)(DESC_PAYLOAD(desc)) +
		offsetof(struct carl9170_tx_superframe, f)));

	desc->dataSize -= sizeof(struct carl9170_tx_superdesc);
	desc->totalLen -= sizeof(struct carl9170_tx_superdesc);
}

static inline void unhide_super(struct dma_desc *desc)
{
	desc->dataAddr = (uint8_t *) get_super(desc);
	desc->dataSize += sizeof(struct carl9170_tx_superdesc);
	desc->totalLen += sizeof(struct carl9170_tx_superdesc);
}

void wlan_tx(struct dma_desc *desc);
void wlan_timer(void);
void wlan_mac_reset(void);
void handle_wlan(void);
void wlan_tx_stuck(const struct ar9170_cmd *cmd, struct ar9170_rsp *rsp);

static inline void __check_wlantx(void)
{
	BUILD_BUG_ON(sizeof(struct ar9170_tx_hwdesc) != AR9170_TX_HWDESC_LEN);
	BUILD_BUG_ON(sizeof(struct carl9170_tx_superdesc) != AR9170_TX_SUPERDESC_LEN);
	BUILD_BUG_ON((offsetof(struct carl9170_tx_superframe, f) & 3) != 0);
	BUILD_BUG_ON(sizeof(struct ar9170_rx_head) != 12);
	BUILD_BUG_ON(sizeof(struct ar9170_rx_phystatus) != 20);
	BUILD_BUG_ON(sizeof(struct ar9170_rx_macstatus) != 4);
}

#endif /* __WLAN_H */
