/*
 * carl9170 firmware - used by the ar9170 wireless device
 *
 * PHY and RF functions
 *
 * Copyright (c) 2000-2005 ZyDAS Technology Corporation
 * Copyright (c) 2007-2009 Atheros Communications, Inc.
 * Copyright	2009	Johannes Berg <johannes@sipsolutions.net>
 * Copyright	2009	Christian Lamparter <chunkeey@googlemail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "config.h"
#include "compiler.h"
#include "types.h"
#include "carl9170.h"
#include "time.h"
#include "printf.h"
#include "phy.h"
#include "linux/phy.h"

#ifdef CONFIG_CARL9170FW_RADIO_FUNCTIONS
static void set_channel_start(void)
{
	/* Manipulate CCA threshold to stop transmission */
	set(AR9170_PHY_REG_CCA_THRESHOLD, 0x300);
	/* Enable Virtual CCA */
	set(AR9170_MAC_REG_VIRTUAL_CCA, 0xf8000);
}

static void set_channel_end(void)
{
	/* Manipulate CCA threshold to resume transmission */
	set(AR9170_PHY_REG_CCA_THRESHOLD, 0x0);
	/* Disable Virtual CCA */
	set(AR9170_MAC_REG_VIRTUAL_CCA, 0x0);
}

void notify_set_channel(void)
{
	set_channel_start();
}

/*
 * Update delta slope coeff man and exp
 */
static void hw_turn_off_dyn(const uint32_t delta_slope_coeff_exp,
			    const uint32_t delta_slope_coeff_man,
			    const uint32_t delta_slope_coeff_exp_shgi,
			    const uint32_t delta_slope_coeff_man_shgi)
{
	uint32_t tmp;

	tmp = get_async(AR9170_PHY_REG_TIMING3) & 0x00001fff;
	tmp |= (delta_slope_coeff_man << AR9170_PHY_TIMING3_DSC_MAN_S) &
		AR9170_PHY_TIMING3_DSC_MAN;
	tmp |= (delta_slope_coeff_exp << AR9170_PHY_TIMING3_DSC_EXP_S) &
		AR9170_PHY_TIMING3_DSC_EXP;

	set(AR9170_PHY_REG_TIMING3, tmp);

	tmp = (delta_slope_coeff_man_shgi << AR9170_PHY_HALFGI_DSC_MAN_S) &
		AR9170_PHY_HALFGI_DSC_MAN;

	tmp |= (delta_slope_coeff_exp_shgi << AR9170_PHY_HALFGI_DSC_EXP_S) &
		AR9170_PHY_HALFGI_DSC_EXP;

	set(AR9170_PHY_REG_HALFGI, tmp);
}

static void program_ADDAC(void)
{
	/* ??? Select Internal ADDAC ??? (is external radio) */
	set(AR9170_PHY_REG_ADC_SERIAL_CTL, AR9170_PHY_ADC_SCTL_SEL_EXTERNAL_RADIO);

	delay(10);

	set(0x1c589c, 0x00000000);	/*# 7-0 */
	set(0x1c589c, 0x00000000);	/*# 15-8 */
	set(0x1c589c, 0x00000000);	/*# 23-16 */
	set(0x1c589c, 0x00000000);	/*# 31- */

	set(0x1c589c, 0x00000000);	/*# 39- */
	set(0x1c589c, 0x00000000);	/*# 47- */
	set(0x1c589c, 0x00000000);	/*# 55- [48]:doubles the xtalosc bias current */
	set(0x1c589c, 0x00000000);	/*# 63- */

	set(0x1c589c, 0x00000000);	/*# 71- */
	set(0x1c589c, 0x00000000);	/*# 79- */
	set(0x1c589c, 0x00000000);	/*# 87- */
	set(0x1c589c, 0x00000000);	/*# 95- */

	set(0x1c589c, 0x00000000);	/*# 103- */
	set(0x1c589c, 0x00000000);	/*# 111- */
	set(0x1c589c, 0x00000000);	/*# 119- */
	set(0x1c589c, 0x00000000);	/*# 127- */

	set(0x1c589c, 0x00000000);	/*# 135- */
	set(0x1c589c, 0x00000000);	/*# 143- */
	set(0x1c589c, 0x00000000);	/*# 151- */
	set(0x1c589c, 0x00000030);	/*# 159- #[158:156]=xlnabufmode */

	set(0x1c589c, 0x00000004);	/*# 167-  [162]:disable clkp_driver to flow */
	set(0x1c589c, 0x00000000);	/*# 175- */
	set(0x1c589c, 0x00000000);	/*# 183-176 */
	set(0x1c589c, 0x00000000);	/*# 191-184 */

	set(0x1c589c, 0x00000000);	/*# 199- */
	set(0x1c589c, 0x00000000);	/*# 207- */
	set(0x1c589c, 0x00000000);	/*# 215- */
	set(0x1c589c, 0x00000000);	/*# 223- */

	set(0x1c589c, 0x00000000);	/*# 231- */
	set(0x1c58c4, 0x00000000);	/*# 233-232 */

	delay(10);

	/* Select External Flow ???? (is internal addac??) */
	set(AR9170_PHY_REG_ADC_SERIAL_CTL, AR9170_PHY_ADC_SCTL_SEL_INTERNAL_ADDAC);
}

static uint32_t AGC_calibration(uint32_t loop)
{
	uint32_t wrdata;
	uint32_t ret;

#define AGC_CAL_NF	(AR9170_PHY_AGC_CONTROL_CAL | AR9170_PHY_AGC_CONTROL_NF);

	wrdata = get_async(AR9170_PHY_REG_AGC_CONTROL) | AGC_CAL_NF;
	set(AR9170_PHY_REG_AGC_CONTROL, wrdata);

	ret = get_async(AR9170_PHY_REG_AGC_CONTROL) & AGC_CAL_NF;

	/* sitesurvey : 100 ms / current connected 200 ms */
	while (loop && ret != 0x0) {
		ret = get_async(AR9170_PHY_REG_AGC_CONTROL) & AGC_CAL_NF;

		if (ret == 0)
			break;

		udelay(100);
		loop--;
	}

	/* return the AGC/Noise calibration state to the driver */
	return ret;
}

static uint32_t init_rf(const uint32_t delta_slope_coeff_exp,
			const uint32_t delta_slope_coeff_man,
			const uint32_t delta_slope_coeff_exp_shgi,
			const uint32_t delta_slope_coeff_man_shgi,
			const uint32_t finiteLoopCount,
			const bool initialize)
{
	uint32_t ret;

	hw_turn_off_dyn(delta_slope_coeff_exp,
			delta_slope_coeff_man,
			delta_slope_coeff_exp_shgi,
			delta_slope_coeff_man_shgi);

	if (initialize) {
		/* Real Chip */
		program_ADDAC();

		/* inverse chain 0 <-> chain 2 */
		set(AR9170_PHY_REG_ANALOG_SWAP, AR9170_PHY_ANALOG_SWAP_AB);

		/* swap chain 0 and chain 2 */
		set(AR9170_PHY_REG_ANALOG_SWAP, AR9170_PHY_ANALOG_SWAP_AB |
						AR9170_PHY_ANALOG_SWAP_ALT_CHAIN);

		/* configure mask */
		set(AR9170_PHY_REG_RX_CHAINMASK, 0x5);	/* chain 0 + chain 2 */
		set(AR9170_PHY_REG_CAL_CHAINMASK, 0x5);	/* chain 0 + chain 2 */

		/* Activate BB */
		set(AR9170_PHY_REG_ACTIVE, AR9170_PHY_ACTIVE_EN);
		delay(10);
	}

	ret = AGC_calibration(finiteLoopCount);

	set_channel_end();
	return ret;
}

void rf_cmd(const struct ar9170_cmd *cmd, struct ar9170_rsp *resp)
{
	uint32_t ret;

	fw.phy.dynamic_ht2040 = cmd->rf_init.dynht2040;
	fw.phy.ht_enable = cmd->rf_init.ht_settings & 1;
	fw.phy.channel_offset = cmd->rf_init.ht_settings >> 2;
	fw.phy.frequency = cmd->rf_init.freq;

	if (fw.phy.dynamic_ht2040 && fw.phy.ht_enable)
		clock_set(true, AHB_80_88MHZ);
	else
		clock_set(true, AHB_40_44MHZ);

	ret = init_rf(cmd->rf_init.delta_slope_coeff_exp,
		      cmd->rf_init.delta_slope_coeff_man,
		      cmd->rf_init.delta_slope_coeff_exp_shgi,
		      cmd->rf_init.delta_slope_coeff_man_shgi,
		      cmd->rf_init.finiteLoopCount,
		      cmd->cmd == AR9170_CMD_RF_INIT);

	resp->len = sizeof(struct ar9170_rf_init_result);
	resp->rf_init_res.ret = ret;

	resp->rf_init_res.regs[0] = get(AR9170_PHY_REG_CCA);
	resp->rf_init_res.regs[3] = get(AR9170_PHY_REG_EXT_CCA);

	resp->rf_init_res.regs[1] = get(AR9170_PHY_REG_CH1_CCA);
	resp->rf_init_res.regs[4] = get(AR9170_PHY_REG_CH1_EXT_CCA);

	resp->rf_init_res.regs[2] = get(AR9170_PHY_REG_CH2_CCA);
	resp->rf_init_res.regs[5] = get(AR9170_PHY_REG_CH2_EXT_CCA);
}

#endif /* CONFIG_CARL9170FW_RADIO_FUNCTIONS */
