/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <error.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "../include/compiler.h"
#include "../include/hdr.h"

#if BYTE_ORDER == LITTLE_ENDIAN
#define CRCPOLY_LE 0xedb88320

/* copied from linux kernel  */
__le32 crc32_le(uint32_t crc, unsigned char const *p, size_t len)
{
        int i;
        while (len--) {
                crc ^= *p++;
                for (i = 0; i < 8; i++)
                        crc = (crc >> 1) ^ ((crc & 1) ? CRCPOLY_LE : 0);
        }
        return crc;
}
#else
#error "this tool does not work with a big endian host yet!"
#endif

const uint8_t otus_magic[4] = { OTUS_MAGIC };

int main(int argc, char* args[])
{
	struct stat stat;
	struct carl9170_fw_otus_desc_v1 *otus_desc = NULL;
	FILE *fw = NULL;
	char *fw_data = NULL;
	size_t fw_len;
	int err, scan = 0, found = 0;
	__le32 crc32;

	if (argc != 2) {
		fprintf(stderr, "usage: %s filename\n", args[0]);
		fprintf(stderr, "This utility can be used on ar9170's firmware file to "
				"apply and verify a crc32 checksum.\n");
		goto fail;
	}

	fw = fopen(args[1], "r+");
	if (fw == NULL) {
		fprintf(stderr, "Failed to open file \"%s\" (%d).\n",
			args[1], errno);
		goto fail;
	}

	err = fstat(fileno(fw), &stat);
	if (err) {
		fprintf(stderr, "Failed to query file infos from \"%s\" (%d).\n",
			args[1], errno);
		goto fail;
	}

	fw_len = stat.st_size;
	if (fw_len < 1024 || fw_len > 32768) {
		fprintf(stderr, "Firmware \"%s\" filesize is out of bound. (%ld).\n",
			args[1], fw_len);
		goto fail;
	}

	fw_data = malloc(fw_len + 4);
	if (fw_data == NULL) {
		fprintf(stderr, "Failed to alloc temporary storage for firmware image.\n");
		goto fail;
	}

	err = fread(fw_data, fw_len, 1, fw);
	if (err != 1) {
		fprintf(stderr, "failed to read from firmware \"%s\" (%d).\n", args[1], errno);
		goto fail;
	}

	while (scan < fw_len) {
		if (fw_data[scan++] == otus_magic[found])
			found++;
		else
			found = 0;

		if (found == sizeof(otus_magic))
			break;
	}

	if (found != sizeof(otus_magic)) {
		fprintf(stderr, "firmware \"%s\" has no descriptor magic.\n", args[1]);
		goto fail;
	}

	otus_desc = (void *) &fw_data[scan - found];

	if (CHECK_HDR_VERSION(&otus_desc->head, CARL9170_FW_OTUS_DESC_CUR_VER)) {
		fprintf(stderr, "Unsupported otus header revision.\n");
		goto fail;
	}

	if (is_feature_supported(otus_desc->fw_feature_set,
	    CARL9170_FW_CRC32_CHECKSUM_TAIL) != 0) {
		crc32 = crc32_le(~0, (void *) fw_data, fw_len - 4);
		if (memcmp(&crc32, &fw_data[fw_len - 4], 4) == 0) {
			fprintf(stderr, "firmware checksum verified.\n");
			goto good_out;
		} else {
			fprintf(stderr, "crc32 mismatch! correcting...\n");
			fw_len -= 4;
		}
	}
	otus_desc->fw_feature_set |= cpu_to_le32(BIT(CARL9170_FW_CRC32_CHECKSUM_TAIL));

	crc32 = crc32_le(~0, (void *) fw_data, fw_len);
	memcpy(&fw_data[fw_len], &crc32, sizeof(crc32));
	rewind(fw);

	err = fwrite(fw_data, fw_len + 4, 1, fw);
	if (err != 1) {
		fprintf(stderr, "failed to write firmware \"%s\" (%d).\n", args[1], errno);
		goto fail;
	}

good_out:
	fprintf(stdout, "firmware's CRC: %x\n", crc32);
	free(fw_data);
	fclose(fw);
	return EXIT_SUCCESS;

fail:
	if (fw_data)
		free(fw_data);

	if (fw)
		fclose(fw);

	return EXIT_FAILURE;
}
