/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <error.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "../include/compiler.h"
#include "../include/hdr.h"

struct feature_list {
	unsigned int id;
	char name[64];
	void (*func)(const struct carl9170_fw_desc_head *, const void *, const size_t);
};

#define CHECK_FOR_FEATURE(feature_enum)					\
	{ .id = feature_enum, .name = #feature_enum, .func = NULL }

#define CHECK_FOR_FEATURE_FUNC(feature_enum, _func)			\
	{ .id = feature_enum, .name = #feature_enum, .func = _func }

static void show_crc32_checksum(const struct carl9170_fw_desc_head *head,
				const void *fw_data, const size_t fw_len)
{
	const uint8_t *raw = fw_data;
	fprintf(stdout, "\t\t\tCRC32: %x\n", le32_to_cpu(*(__le32 *)&raw[fw_len - 4]));
}

static const struct feature_list known_otus_features_v1[] = {
	CHECK_FOR_FEATURE(CARL9170_FW_DUMMY_FEATURE),
	CHECK_FOR_FEATURE(CARL9170_FW_INTEGRATED_DESC_HEADER),
	CHECK_FOR_FEATURE_FUNC(CARL9170_FW_CRC32_CHECKSUM_TAIL, show_crc32_checksum),
	CHECK_FOR_FEATURE(CARL9170_FW_UNUSEABLE),
	CHECK_FOR_FEATURE(AR9170_FW_COMMAND_BASIC),
	CHECK_FOR_FEATURE(AR9170_FW_COMMAND_PHY),
	CHECK_FOR_FEATURE(AR9170_FW_COMMAND_CAM),
	CHECK_FOR_FEATURE(CARL9170_FW_WLANTX_CAB),
	CHECK_FOR_FEATURE(AR9170_FW_WLANTX_DELAYED_TX),
	CHECK_FOR_FEATURE(AR9170_FW_HANDLE_BACK_REQ),
	CHECK_FOR_FEATURE(AR9170_FW_LEGACY),
	CHECK_FOR_FEATURE(CARL9170_FW_CARL9170_V1),
};

static void check_feature_list(const struct carl9170_fw_desc_head *head,
			       const __le32 bitmap,
			       const struct feature_list *list,
			       const unsigned int entries,
			       const void *fw_data, const size_t fw_len)
{
	unsigned int i;

	for (i = 0; i < entries; i++) {
		if (!is_feature_supported(bitmap, list[i].id))
			continue;

		fprintf(stdout, "\t\t%2d = %s\n", list[i].id, list[i].name);
		if (list[i].func)
			list[i].func(head, fw_data, fw_len);
	}
}

static void show_otus_desc(const struct carl9170_fw_desc_head *head,
			   const void *fw_data, const size_t fw_len)
{
	const struct carl9170_fw_otus_desc_v1 *otus = (const void *) head;
	unsigned int i;
	unsigned int fw_date;

	BUILD_BUG_ON(ARRAY_SIZE(known_otus_features_v1) != __CARL9170_FW_FEATURE_NUM);

	fw_date = otus->fw_year_month_day;
	fprintf(stdout, "\tFirmware Build Date (YYYY-MM-DD): 2%03d-%02d-%02d\n",
		CARL9170_FW_GET_YEAR(fw_date), CARL9170_FW_GET_MONTH(fw_date),
		CARL9170_FW_GET_DAY(fw_date));

	fprintf(stdout, "\tBeacon Address: %x, (reserved length:%d)\n",
		le32_to_cpu(otus->bcn_addr), le16_to_cpu(otus->bcn_len));

	fprintf(stdout, "\tSupported Features: (raw:%.08x)\n",
		le32_to_cpu(otus->fw_feature_set));

	check_feature_list(head, otus->fw_feature_set, known_otus_features_v1,
			   ARRAY_SIZE(known_otus_features_v1), fw_data, fw_len);
}

static void show_miniboot_info(const struct carl9170_fw_desc_head *head,
			       const void *fw_data, const size_t fw_len)
{
	const struct carl9170_fw_usb_desc_v1 *usb = (const void *) head;

	fprintf(stdout, "\t\t\tminiboot size: %d\n", usb->miniboot_size);
}

static const struct feature_list known_usb_features_v1[] = {
	CHECK_FOR_FEATURE(CARL9170_FW_USB_DUMMY_FEATURE),
	CHECK_FOR_FEATURE_FUNC(CARL9170_FW_USB_MINIBOOT, show_miniboot_info),
	CHECK_FOR_FEATURE(AR9170_FW_USB_INIT_FIRMWARE),
	CHECK_FOR_FEATURE(AR9170_FW_USB_RESP_EP2),
	CHECK_FOR_FEATURE(AR9170_FW_USB_DOWN_STREAM),
	CHECK_FOR_FEATURE(AR9170_FW_USB_UP_STREAM),
};

static void show_usb_desc(const struct carl9170_fw_desc_head *head,
			  const void *fw_data, const size_t fw_len)
{
	const struct carl9170_fw_usb_desc_v1 *usb = (const void *) head;
	unsigned int i;

	BUILD_BUG_ON(ARRAY_SIZE(known_usb_features_v1) != __CARL9170_FW_USB_FEATURE_NUM);

	fprintf(stdout, "\tTX DMA chunk size:%d Bytes, TX DMA chunks:%d\n",
		usb->tx_frag_len, usb->tx_descs);
	fprintf(stdout, "\t=> %d Bytes are reserved for the TX queues\n",
		usb->tx_frag_len * usb->tx_descs);
	fprintf(stdout, "\tMaximum RX Stream block:%d Bytes\n",
		usb->rx_max_frame_len);

	fprintf(stdout, "\tSupported Features: (raw:%.08x)\n",
		le32_to_cpu(usb->usb_feature_set));

	check_feature_list(head, usb->usb_feature_set, known_usb_features_v1,
			   ARRAY_SIZE(known_usb_features_v1), fw_data, fw_len);
}

static void show_motd_desc(const struct carl9170_fw_desc_head *head,
			   const void *fw_data, const size_t fw_len)
{
	const struct carl9170_fw_motd_desc_v1 *motd = (const void *) head;
	char buf[CARL9170_FW_MOTD_STRING_LEN];

	strncpy(buf, motd->desc, CARL9170_FW_MOTD_STRING_LEN);
	fprintf(stdout, "\tFirmware Description:\"%s\"\n", buf);

	strncpy(buf, motd->release, CARL9170_FW_MOTD_STRING_LEN);
	fprintf(stdout, "\tFirmware Release:\"%s\"\n", buf);
}

static void show_fix_desc(const struct carl9170_fw_desc_head *head,
			  const void *fw_data, const size_t fw_len)
{
	const struct carl9170_fw_fix_desc_v1 *fix = (const void *) head;
	const struct carl9170_fw_fix_entry *iter;
	int i;

	for (i = 0; i < (head->length - sizeof(*head)) / sizeof(*iter); i++) {
		iter = &fix->data[i];
		fprintf(stdout, "\t %d: 0x%.x8 := 0x%.x8 (0x%.x8)\n", i,
			le32_to_cpu(iter->address),
			le32_to_cpu(iter->value),
			le32_to_cpu(iter->mask));
	}
}

static void show_debug_desc(const struct carl9170_fw_desc_head *head,
			    const void *fw_data, const size_t fw_len)
{
	const struct carl9170_fw_debug_desc_v1 *dbg = (const void *) head;

	/* Nothing interesting here */
}

static void show_last_desc(const struct carl9170_fw_desc_head *head,
			   const void *fw_data, const size_t fw_len)

{
	const struct carl9170_fw_last_desc *last = (const void *) head;
}

#define ADD_HANDLER(_magic, _min_ver, _func)			\
	{ .magic = _magic, .min_ver = _min_ver, .func = _func }

static const struct {
	uint8_t magic[4];
	uint8_t min_ver;
	void (*func)(const struct carl9170_fw_desc_head *, const void *, const size_t);
} known_magics[] = {
	ADD_HANDLER(OTUS_MAGIC, CARL9170_FW_OTUS_DESC_CUR_VER, show_otus_desc),
	ADD_HANDLER(USB_MAGIC, CARL9170_FW_USB_DESC_CUR_VER, show_usb_desc),
	ADD_HANDLER(MOTD_MAGIC, CARL9170_FW_MOTD_DESC_CUR_VER, show_motd_desc),
	ADD_HANDLER(DBG_MAGIC, CARL9170_FW_MOTD_DESC_CUR_VER, show_debug_desc),
	ADD_HANDLER(FIX_MAGIC, CARL9170_FW_MOTD_DESC_CUR_VER, show_fix_desc),
	ADD_HANDLER(LAST_MAGIC, CARL9170_FW_LAST_DESC_CUR_VER, show_last_desc),
};

static const uint8_t otus_magic[4] = { OTUS_MAGIC };

#define SHOW_DESC_HEAD(h)							\
	fprintf(stdout, "%c%c%c%c Descriptor: length:%d, compatible:%d, "	\
		"version:%d\n",	h->magic[0], h->magic[1], h->magic[2],		\
		h->magic[3], le16_to_cpu(h->length), h->min_ver, h->cur_ver)

int main(int argc, char *args[])
{
	struct stat stat;
	struct carl9170_fw_desc_head *fw_desc;
	struct carl9170_fw_otus_desc_v1 *otus_desc = NULL;
	FILE *fw = NULL;
	char *fw_data = NULL;
	size_t fw_len;
	unsigned int i, len = 0;
	int err = -1, scan = 0, found = 0, extra;

	if (argc != 2) {
		fprintf(stderr, "usage: %s filename.fw\n", args[0]);
		fprintf(stderr, "This utility can be used on ar9170's firmware file to "
				"display all known flags.\n");
		goto out;
	}

	fw = fopen(args[1], "r+");
	if (fw == NULL) {
		fprintf(stderr, "Failed to open file \"%s\" (%d).\n", args[1], errno);
		goto out;
	}

	err = fstat(fileno(fw), &stat);
	if (err) {
		fprintf(stderr, "File \"%s\" not accessible (%d).\n", args[1], errno);
		goto out;
	}

	fw_len = stat.st_size;
	if (fw_len < 1024 || fw_len > 32768) {
		fprintf(stderr, "Firmware \"%s\" size is out of bound. (%ld).\n",
			args[1], fw_len);
		goto out;
	}

	fprintf(stdout, "Total firmware size: %d Bytes\n\n", fw_len);

	fw_data = malloc(fw_len);
	if (fw_data == NULL) {
		fprintf(stderr, "Failed to alloc temporary storage for image.\n");
		goto out;
	}

	err = fread(fw_data, fw_len, 1, fw);
	if (err != 1) {
		fprintf(stderr, "Failed to read from \"%s\" (%d).\n", args[1], errno);
		goto out;
	}

	while (scan < (fw_len - 512)) {
		if (fw_data[scan] == otus_magic[found]) {
			if (++found == sizeof(otus_magic)) {
				otus_desc = (void *) &fw_data[scan - found + 1];
				break;
			}
		} else {
			found = 0;
		}

		scan++;
	}

	if (!otus_desc) {
		fprintf(stderr, "firmware \"%s\" has not all necessary "
				"descriptors.\n", args[1]);
		goto out;
	}

	for_each_hdr(fw_desc, &otus_desc->head) {
		len += fw_desc->length;
	}
	len += fw_desc->length;

	fprintf(stdout, "Firmware Header at position:%d, length: %d Bytes\n\n",
		scan - found + 1, len);

	if (len > CARL9170_FW_DESC_MAX_LENGTH) {
		fprintf(stderr, "Firmware header is too long!\n");
		goto out;
	}

	for_each_hdr(fw_desc, &otus_desc->head) {
		SHOW_DESC_HEAD(fw_desc);
		len += fw_desc->length;

		for (i = 0; i < ARRAY_SIZE(known_magics); i++) {
			if (memcmp(fw_desc->magic, known_magics[i].magic, 4) == 0) {
				if (CHECK_HDR_VERSION(fw_desc, known_magics[i].min_ver))
					fprintf(stdout, "Unsupported header revision\n");
				else
					known_magics[i].func(fw_desc, fw_data, fw_len);
				break;
			}
		}

		if (i == ARRAY_SIZE(known_magics))
			fprintf(stdout, "unknown descriptor\n");

		fprintf(stdout, "\n");
	}
	SHOW_DESC_HEAD(fw_desc);

out:
	if (fw_data)
		free(fw_data);

	if (fw)
		fclose(fw);

	return err ? EXIT_FAILURE : EXIT_SUCCESS;
}
