/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <error.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "../include/compiler.h"
#include "../include/hdr.h"

const uint8_t otus_magic[4] = { OTUS_MAGIC };
const uint8_t usb_magic[4] = { USB_MAGIC };

int main(int argc, char *args[])
{
	struct stat stat;
	struct carl9170_fw_otus_desc_v1 *otus_desc = NULL;
	struct carl9170_fw_usb_desc_v1 *usb_desc = NULL;
	FILE *fw = NULL, *m = NULL;
	char *buf = NULL;
	size_t extra, fw_len;
	int err, scan = 0, found = 0;

	if (argc != 3) {
		fprintf(stderr, "usage: %s filename.fw miniboot.fw\n", args[0]);
		fprintf(stderr, "This utility can be used on ar9170's firmware file to "
				"make it incompatible with old drivers.\n");
		goto fail;
	}

	fw = fopen(args[1], "r+");
	if (fw == NULL) {
		fprintf(stderr, "Failed to open file \"%s\" (%d).\n",
			args[1], errno);
		goto fail;
	}

	m = fopen(args[2], "r");
	if (m == NULL) {
		fprintf(stderr, "Failed to open file %s (%d).\n",
			args[2], errno);
		goto fail;
	}

	err = fstat(fileno(m), &stat);
	if (err) {
		fprintf(stderr, "Failed to query file infos from \"%s\" (%d).\n",
			args[2], errno);
		goto fail;
	}

	extra = stat.st_size;

	err = fstat(fileno(fw), &stat);
	if (err) {
		fprintf(stderr, "Failed to query file infos from %s (%d).\n",
			args[1], errno);
		goto fail;
	}

	fw_len = stat.st_size;
	if (fw_len < 1024 || fw_len > 32768) {
		fprintf(stderr, "Firmware \"%s\" filesize is out of bound. (%ld).\n",
			args[1], stat.st_size);
		goto fail;
	}

	buf = malloc(fw_len + extra);
	if (buf == NULL) {
		fprintf(stderr, "Failed to alloc temporary storage for firmware image.\n");
		goto fail;
	}

	err = fread(buf, extra, 1, m);
	if (err != 1) {
		fprintf(stderr, "Failed to read from miniboot \"%s\" (%d).\n",
			args[2], errno);
		goto fail;
	}

	err = fread(&buf[extra], fw_len, 1, fw);
	if (err != 1) {
		fprintf(stderr, "Failed to read from firmware \"%s\" (%d).\n",
			args[1], errno);
		goto fail;
	}

	while (scan < fw_len + extra - 512) {
		if (!otus_desc && buf[scan] == otus_magic[found]) {
			if (++found == sizeof(otus_magic)) {
				otus_desc = (void *) &buf[scan - found + 1];
				found = 0;
				scan++;
				continue;
			}
		} else if (!usb_desc && buf[scan] == usb_magic[found]) {
			if (++found == sizeof(usb_magic)) {
				usb_desc = (void *) &buf[scan - found + 1];
				found = 0;
				scan++;
				continue;
			}
		} else
			found = 0;

		if (otus_desc && usb_desc)
			break;

		scan++;
	}

	if (!otus_desc || !usb_desc) {
		fprintf(stderr, "Firmware \"%s\" has not all necessary "
				"descriptors.\n", args[1]);
		goto fail;
	}

	if (CHECK_HDR_VERSION(&otus_desc->head, CARL9170_FW_OTUS_DESC_CUR_VER) ||
	    CHECK_HDR_VERSION(&usb_desc->head, CARL9170_FW_USB_DESC_CUR_VER)) {
		fprintf(stderr, "Unsupported firmware header revisions.\n");
		goto fail;
	}

	if (is_feature_supported(otus_desc->fw_feature_set, AR9170_FW_LEGACY)) {
		fprintf(stderr, "Firmware \"%s\" is supposed to be compatible "
			"with legacy drivers.\n", args[1]);
		goto fail;
	}

	if (is_feature_supported(otus_desc->fw_feature_set, CARL9170_FW_CRC32_CHECKSUM_TAIL)) {
		fprintf(stderr, "Firmware \"%s\" is already protected by CRC32.\n",
			args[1]);
		goto fail;
	}

	if (is_feature_supported(usb_desc->usb_feature_set, CARL9170_FW_USB_MINIBOOT)) {
		fprintf(stderr, "Firmware \"%s\" has already a miniboot image.\n",
			args[1]);
		goto fail;
	}

	usb_desc->usb_feature_set |= cpu_to_le32(BIT(CARL9170_FW_USB_MINIBOOT));
	usb_desc->miniboot_size = cpu_to_le16(extra);

	rewind(fw);

	err = fwrite(buf, stat.st_size + extra, 1, fw);
	if (err != 1) {
		fprintf(stderr, "Failed to write firmware \"%s\" (%d).\n", args[1], errno);
		goto fail;
	}

	free(buf);
	fclose(m);
	fclose(fw);

	fprintf(stdout, "Miniboot applied.\n");
	return EXIT_SUCCESS;

fail:
	if (buf)
		free(buf);

	if (fw)
		fclose(fw);
	if (m)
		fclose(m);

	return EXIT_FAILURE;
}
